use serde_json;
use serde::Serialize;
use failure::Fail;

#[derive(Fail, Debug)]
pub enum UserJourneyError
{
    #[fail(display = "The JSON String could not be created. ({})", _0)]
    SerializationFailed(serde_json::error::Error)
}

#[derive(Serialize)]
pub struct UserJourney
{
    // file_name: String,
    pub origin_country: Option<String>,
    pub destination_country: Option<String>,
    pub workflow_type: Option<String>,
    pub description: Vec<LocaleValue>,
    pub phases: Vec<Phase>
}

impl UserJourney
{
    pub fn new_empty() -> UserJourney
    {
        UserJourney
        {
            origin_country: None,
            destination_country: None,
            workflow_type: None,
            description: Vec::new(),
            phases: Vec::new()
        }
    }

    pub fn is_concrete(&self) -> bool
    {
        self.origin_country.is_some() 
            && self.destination_country.is_some() 
            && self.workflow_type.is_some()
            && !self.get_origin_country().unwrap_or_default().trim().eq("")
            && !self.get_destination_country().unwrap_or_default().trim().eq("")
            && !self.get_workflow_type().unwrap_or_default().trim().eq("")
            && self.has_all_touchpoints()
            && self.has_all_english_values()
    }

    pub fn has_all_english_values(&self) -> bool
    {
        for phase in self.phases.iter()
        {
            for name in phase.get_name().iter()
            {
                if name.locale == "en" && name.value.trim() == ""
                {
                    return false;
                }
            }
            
            for action in phase.actions.iter()
            {
                for name in action.get_name().iter()
                {
                    if name.locale == "en" && name.value.trim() == ""
                    {
                        return false;
                    }
                }

                for description in action.get_description().iter()
                {
                    if description.locale == "en" && description.value.trim() == ""
                    {
                        return false;
                    }
                }

            }
        }
        for description in self.description.iter()
        {
            if description.locale == "en" && description.value.trim() == ""
            {
                return false;
            }
        }
        return true;
    }

    pub fn has_all_touchpoints(&self) -> bool
    {
        for phase in self.phases.iter()
        {
            for action in phase.actions.iter()
            {
                if action.touchpoints.is_empty()
                {
                    return false;
                }
            }
        }
        return true;
    }

    pub fn set_origin_country(&mut self, origin_country : Option<String>) 
    {
        self.origin_country = origin_country;
    }

    pub fn set_destination_country(&mut self, destination_country : Option<String>) 
    {
        self.destination_country = destination_country;
    }

    pub fn set_user_journey_description(&mut self, description : Vec<LocaleValue>) 
    {
        self.description = description;
    }

    pub fn set_workflow_type(&mut self, workflow_type : Option<String>) 
    {
        self.workflow_type = workflow_type;
    }

    pub fn get_workflow_type(&self) -> Option<String> {
        if let Some(workflow_type) = &self.workflow_type {
            return Some(workflow_type.to_string());
        } else {
            return None;
        }
    }

    pub fn get_origin_country(&self) -> Option<String> {
        if let Some(origin_country) = &self.origin_country {
            return Some(origin_country.to_string());
        } else {
            return None;
        }
    }

    pub fn get_destination_country(&self) -> Option<String> {
        if let Some(destination_country) = &self.destination_country {
            return Some(destination_country.to_string());
        } else {
            return None;
        }
    }

    pub fn contains_phase(&self, drawio_id: &str) -> bool {
        for phase in self.phases.iter()
        {
            if phase.get_drawio_id().eq(drawio_id) {
                return true;
            }
        }
        return false;
    }

    pub fn add_phase(&mut self, phase: Phase) {
        self.phases.push(phase)
    }

}

#[derive(Serialize)]
pub struct Phase
{
    order: u32,
    drawio_id: String,
    name: Vec<LocaleValue>,
    pub actions: Vec<Action>
}

impl Phase
{
    pub fn new(drawio_id: String, index: u32, name: Vec<LocaleValue>, actions: Vec<Action>) -> Phase
    {
        Phase
        {
            order: index,
            drawio_id: drawio_id,
            name: name,
            actions: actions
        }
    }

    pub fn get_drawio_id(&self) -> String {
        return self.drawio_id.to_string();
    }

    pub fn get_name(&self) -> &Vec<LocaleValue> {
        return &self.name;
    }

    pub fn add_action(&mut self, action: Action) {
        self.actions.push(action)
    }
}

#[derive(Serialize)]
pub struct Action
{
    order: u32,
    pub drawio_id: String,
    pub name: Vec<LocaleValue>,
    pub description: Vec<LocaleValue>,
    pub touchpoints: Vec<Touchpoint>,
    pub is_mandatory: bool
}

impl Action
{
    pub fn new_optional(drawio_id: String, index: u32, name: Vec<LocaleValue>, description: Vec<LocaleValue>,
        touchpoints: Vec<Touchpoint>) -> Action
    {
        Action
        {
            order: index,
            drawio_id: drawio_id,
            name: name,
            description: description,
            touchpoints: touchpoints,
            is_mandatory: false
        }
    }

    pub fn new_mandatory(drawio_id: String, index: u32, name: Vec<LocaleValue>, description: Vec<LocaleValue>,
        touchpoints: Vec<Touchpoint>) -> Action
    {
        Action
        {
            order: index,
            drawio_id: drawio_id,
            name: name,
            description: description,
            touchpoints: touchpoints,
            is_mandatory: true
        }
    }

    pub fn add_touchpoints(&mut self, touchpoint: Touchpoint) {
        self.touchpoints.push(touchpoint)
    }

    pub fn get_drawio_id(&self) -> String {
        return self.drawio_id.to_string();
    }

    pub fn get_name(&self) -> &Vec<LocaleValue> {
        return &self.name;
    }

    pub fn get_description(&self) -> &Vec<LocaleValue> {
        return &self.description;
    }

}

#[derive(Serialize)]
pub struct Touchpoint
{
    pub drawio_id: String,
    pub service_id: String,
    service_url: String,
    pub location: String,
    name: Option<String>,
    description: Option<String>
}

impl Touchpoint
{
    pub fn new(drawio_id: String, service_id: String, service_url: String, location: String, 
        name: Option<String>, description: Option<String>) -> Touchpoint
    {
        Touchpoint
        {
            drawio_id: drawio_id,
            service_id: service_id,
            service_url: service_url,
            location: location,
            name: name,
            description: description
        }
    }

    pub fn get_drawio_id(&self) -> String {
        return self.drawio_id.to_string();
    }
}

#[derive(Serialize)]
pub struct LocaleValue
{
    locale: String,
    value: String
}

impl LocaleValue
{
    pub fn new(locale: String, value: String) -> LocaleValue
    {
        LocaleValue
        {
            locale: locale,
            value: value
        }
    }

}

pub fn serialize(user_journey: UserJourney) -> Result<String, UserJourneyError>{
    let user_journey = serde_json::to_string(&user_journey)
        .map_err(|error| UserJourneyError::SerializationFailed(error))?;
    Ok(user_journey)
}