use super::UseCaseExecError;
use crate::entities::{bpmn, bpmn::datatypes::{tasks::{service_task::ServiceTask, sub_process::SubProcess},
    flows::sequence_flow::SequenceFlow, bpmndi_bpmn_diagram::{DiBPMNShape, DiBPMNEdge}}};
use crate::domain::model;

pub fn from_user_journey(user_journey: &model::UserJourney) -> Result<String, UseCaseExecError>
{
    // Set initial BPMN information
    let process_id = "def_process".to_string();
    let process_name = "Default Process".to_string();

    // Create initial BPMN definitions
    let mut definitions = bpmn::Definitions::init(process_id, Some(process_name));

    // Create and set Start Event and Shape
    let start_event = bpmn::datatypes::events::start_event::StartEvent::init(bpmn::config::DEF_START_EVENT_ID.to_string(), bpmn::config::DEF_START_EVENT_NAME.to_string());
    definitions.process.start_event = Some(start_event);
    let start_event_shape = DiBPMNShape::new_init_event(bpmn::config::DEF_START_EVENT_ID.to_string(), bpmn::config::DEF_START_EVENT_X_POS.to_string());
    definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(start_event_shape);
    
    // Create and set Tasks
    let mut prev_element_id = bpmn::config::DEF_START_EVENT_ID.to_string();
    let mut xs : u32 = bpmn::config::DEF_EVENT_WIDTH;
    let mut xt : u32;

    for phase in user_journey.phases.iter()
    {
        // Create new Sub Process
        let sub_process_id = format!{"sub_process_{}", phase.get_drawio_id()};
        let mut sub_process = SubProcess::new_initial(sub_process_id.clone(), phase.get_name());

        // Create Sequence Flow to new Sub Process
        let flow_id = format!{"sequence_flow_{}", phase.get_drawio_id()};
        let flow = SequenceFlow::new_unnamed(flow_id.clone(), prev_element_id.to_string(), sub_process_id.clone());
        definitions.process.push_to_flows(flow);

        // Create BPMN Edge for new Sequence Flow
        xt = xs + bpmn::config::DEF_SEQUENCE_FLOW_SIZE;
        let edge = DiBPMNEdge::init(flow_id.clone(), xs.to_string(), xt.to_string());
        definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_edges(edge);

        // Remember xt as xs for Sub Process Shape
        let xs_sub_process = xt;

        // Create new Start Event in new Sub Process
        let sub_process_start_event_id = format!{"start_event_{}", sub_process_id.clone()};
        let sub_process_start_event = bpmn::datatypes::events::start_event::StartEvent::init(sub_process_start_event_id.clone(), format!{"{} {}", phase.get_name(), bpmn::config::DEF_START_EVENT_NAME.to_string()});
        sub_process.start_event = Some(sub_process_start_event);

        // Create BPMN Shape for new Start Event
        xs = xt + bpmn::config::DEF_SUB_PROCESS_OFFSET;
        let sub_process_start_event_shape = DiBPMNShape::new_init_event(sub_process_start_event_id.to_string(), xs.to_string());
        definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(sub_process_start_event_shape);
        
        xs = xs + bpmn::config::DEF_EVENT_WIDTH;

        let mut prev_sub_element_id = sub_process_start_event_id;
        // Create a new Service Task for each Action in new Sub Process
        for action in phase.actions.iter()
        {
            // Create new Task
            let task_id = format!{"service_task_{}", action.get_drawio_id()};
            let mut task: ServiceTask;
            if action.is_mandatory {
                task = ServiceTask::new_mandatory(task_id.clone(), action.get_name());
            } else {
                task = ServiceTask::new_optional(task_id.clone(), action.get_name());
            }
            for touchpoint in action.touchpoints.iter()
            {
                task.add_service_id(touchpoint.service_id.to_string());
            }
            sub_process.push_to_tasks(task);

            // Create Sequence Flow to new Service Task
            let flow_id = format!{"sequence_flow_{}", action.get_drawio_id()};
            let flow = SequenceFlow::new_unnamed(flow_id.clone(), prev_sub_element_id.to_string(), task_id.clone());
            sub_process.push_to_flows(flow);

            // Create BPMN Edge for new Sequence Flow
            xt = xs + bpmn::config::DEF_SEQUENCE_FLOW_SIZE;
            let edge = DiBPMNEdge::init(flow_id.clone(), xs.to_string(), xt.to_string());
            definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_edges(edge);

            // Create BPMN Shape for new Service Task
            let shape = DiBPMNShape::new_init_task(task_id.clone(), xt.to_string());
            definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(shape);
            xs = xt + bpmn::config::DEF_TASK_WIDTH;

            // Set Task ID as Prev ID
            prev_sub_element_id = task_id.clone();
        }

        // Create new End Event in Sub Process
        let sub_process_end_event_id = format!{"end_event_{}", sub_process_id.clone()};
        let sub_process_end_event = bpmn::datatypes::events::end_event::EndEvent::init(sub_process_end_event_id.clone(), format!{"{} {}", phase.get_name(), bpmn::config::DEF_END_EVENT_NAME.to_string()});
        sub_process.end_event = Some(sub_process_end_event);

        // Create Sequence Flow to new End Event in Sub Process
        let sub_process_end_event_flow_id = format!{"sequence_flow_{}", sub_process_end_event_id.clone()};
        let sub_process_end_event_flow = SequenceFlow::new_unnamed(sub_process_end_event_flow_id.clone(), prev_sub_element_id, sub_process_end_event_id.clone());
        sub_process.push_to_flows(sub_process_end_event_flow);

        // Add Sub Process to Process in Definitions
        definitions.process.push_to_sub_processes(sub_process);

        // Create BPMN Edge for new Sequence Flow
        xt = xs + bpmn::config::DEF_SEQUENCE_FLOW_SIZE;
        let sub_process_end_event_flow_edge = DiBPMNEdge::init(sub_process_end_event_flow_id.clone(), xs.to_string(), xt.to_string());
        definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_edges(sub_process_end_event_flow_edge);

        // Create BPMN Shape for End Event
        let sub_process_end_event_shape = DiBPMNShape::new_init_event(sub_process_end_event_id.clone(), xt.to_string());
        definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(sub_process_end_event_shape);

        xs = xt + bpmn::config::DEF_EVENT_WIDTH + bpmn::config::DEF_SUB_PROCESS_OFFSET;

        // Create BPMN Shape for new Sub Process
        let shape = DiBPMNShape::new_init_sub_process(sub_process_id.clone(), xs_sub_process.to_string(), (xs - xs_sub_process).to_string());
        definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(shape);

        // Set Task ID as Prev ID
        prev_element_id = sub_process_id.clone();
    }

    // Create and set End Event
    let end_event = bpmn::datatypes::events::end_event::EndEvent::init(bpmn::config::DEF_END_EVENT_ID.to_string(), bpmn::config::DEF_END_EVENT_NAME.to_string());
    definitions.process.end_event = Some(end_event);

    // Create Sequence Flow to End Event
    let flow_id = format!{"sequence_flow_{}", bpmn::config::DEF_END_EVENT_ID.to_string()};
    let flow = SequenceFlow::new_unnamed(flow_id.clone(), prev_element_id, bpmn::config::DEF_END_EVENT_ID.to_string());
    definitions.process.push_to_flows(flow);

    // Create BPMN Edge for new Sequence Flow
    xt = xs + bpmn::config::DEF_SEQUENCE_FLOW_SIZE;
    let edge = DiBPMNEdge::init(flow_id.clone(), xs.to_string(), xt.to_string());
    definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_edges(edge);

    // Create BPMN Shape for End Event
    let end_event_shape = DiBPMNShape::new_init_event(bpmn::config::DEF_END_EVENT_ID.to_string(), xt.to_string());
    definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(end_event_shape);

    /*if actions.is_empty()
    {
        for action in user_journey.actions.iter()
        {
            actions.push(&action);
        }
    }

    for action in actions.iter()
    {
        // Create new Task
        let task_id = format!{"service_task_{}", action.get_drawio_id()};
        let mut task: ServiceTask;
        if action.is_mandatory {
            task = ServiceTask::new_mandatory(task_id.clone(), action.get_name());
        } else {
            task = ServiceTask::new_optional(task_id.clone(), action.get_name());
        }
        
        for touchpoint in action.touchpoints.iter()
        {
            task.add_service_id(touchpoint.service_id.to_string());
        }
        definitions.process.push_to_tasks(task);

        // Create Sequence Flow to new Task
        let flow_id = format!{"sequence_flow_{}", action.drawio_id};
        let flow = SequenceFlow::new_unnamed(flow_id.clone(), prev_element_id.to_string(), task_id.clone());
        definitions.process.push_to_flows(flow);

        // Create BPMN Edge for new Sequence Flow
        xt = xs + bpmn::config::DEF_SEQUENCE_FLOW_SIZE;
        let edge = DiBPMNEdge::init(flow_id.clone(), xs.to_string(), xt.to_string());
        definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_edges(edge);

        // Create BPMN Shape for new Task
        let shape = DiBPMNShape::new_init_task(task_id.clone(), xt.to_string());
        definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(shape);
        xs = xt + bpmn::config::DEF_TASK_WIDTH;

        // Set Task ID as Prev ID
        prev_element_id = task_id.clone();
    }
    
    // Create and set End Event
    let end_event = bpmn::datatypes::events::end_event::EndEvent::init(bpmn::config::DEF_END_EVENT_ID.to_string(), bpmn::config::DEF_END_EVENT_NAME.to_string());
    definitions.process.end_event = Some(end_event);

    // Create Sequence Flow to End Event
    let flow_id = format!{"sequence_flow_{}", bpmn::config::DEF_END_EVENT_ID.to_string()};
    let flow = SequenceFlow::new_unnamed(flow_id.clone(), prev_element_id, bpmn::config::DEF_END_EVENT_ID.to_string());
    definitions.process.push_to_flows(flow);

    // Create BPMN Edge for new Sequence Flow
    xt = xs + bpmn::config::DEF_SEQUENCE_FLOW_SIZE;
    let edge = DiBPMNEdge::init(flow_id.clone(), xs.to_string(), xt.to_string());
    definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_edges(edge);

    // Create BPMN Shape for End Event
    let end_event_shape = DiBPMNShape::new_init_event(bpmn::config::DEF_END_EVENT_ID.to_string(), xt.to_string());
    definitions.bpmndi_bpmn_diagram.push_to_bpmndi_bpmn_shapes(end_event_shape);*/

    let bpmn_string = bpmn::serialize(definitions)
        .map_err(|error| UseCaseExecError::BPMNSerializationFailed(error))?;
    Ok(bpmn_string)
}