use super::UseCaseExecError;
use crate::entities::drawio_xml;
use crate::domain::model;
use std::collections::HashMap;

pub fn from_drawio_xml(mx_graph_model: &drawio_xml::MxGraphModel) -> Result<model::UserJourney, UseCaseExecError>
{
    let mut user_journey = model::UserJourney::new_empty();

    let cells = mx_graph_model.get_mx_cells();

    if cells.is_none() || cells.unwrap().len() < 1
    {
        info!("The User Journey is empty.");
        return Ok(user_journey);
    }

    // Set origin_country, destination_country and/or workflow_type if they are set in the User Journey.
    if let Some(user_journey_properties) = mx_graph_model.get_user_journey_properties()
    {
        if user_journey_properties.has_workflow_properties()
        {
            let mut locale_values = Vec::new();
            for ujlang in user_journey_properties.get_ujlang_list() {

                // include english and all other descriptions with values
                if ujlang.get_value() != "" || ujlang.get_language() == "en" {
                    locale_values.push(model::LocaleValue::new(ujlang.get_language(), ujlang.get_value()))
                }
            }

            user_journey.set_origin_country(user_journey_properties.get_origin_country());
            user_journey.set_destination_country(user_journey_properties.get_destination_country());
            user_journey.set_workflow_type(user_journey_properties.get_workflow_type());
            
            user_journey.set_user_journey_description(locale_values);
        }
    }
    
    // Set Phases, if the User Journey contains some and if they are correctly modeled.
    let edges_by_target = mx_graph_model.get_edge_mx_cells_ref_by_target_map();
    let phases_by_id = mx_graph_model.get_phase_by_id_map();

    let phases_len = phases_by_id.len();
    info!("The User Journey contains {} phase(s).", &phases_len);

    if 0 < phases_len
    {
        let mut start_phases = Vec::new();

        for phase in phases_by_id.values()
        {
            // if a phase has incoming arrows only from non-phase objects, it is a start phase
            if let Some(edges) = edges_by_target.get(&phase.get_id())
            {
                let mut is_start_phase = true;
                for edge in edges
                {
                    if let Some(source) = edge.get_source()
                    {
                        if phases_by_id.contains_key(&source)
                        {
                            is_start_phase = false;
                        }
                    }
                }
                if is_start_phase
                {
                    start_phases.push(phase);
                }
            } else {
            // if a phase has no incoming arrows, it is a start phase
                start_phases.push(phase);
            }
        }

        match start_phases.len()
        {
            0 => Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey with phases must have one start phase.".to_string())),
            1 => 
            {
                let edges_by_source = mx_graph_model.get_edge_mx_cells_ref_by_source_map();
                let actions_by_parent = mx_graph_model.get_actions_by_parent_map();
                let touchpoints_by_parent_action = mx_graph_model.get_touchpoints_by_parent_action_map();
                let descriptions_by_action = mx_graph_model.get_description_by_action_id_map();

                let start_phase = start_phases.get(0).unwrap();
                add_phase(0, start_phase, user_journey, edges_by_source, phases_by_id,
                    actions_by_parent, descriptions_by_action, touchpoints_by_parent_action, mx_graph_model)
            }
            _ => Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey can only have one start phase.".to_string()))
        }
    } else {
        /*let action_by_id = mx_graph_model.get_action_by_id_map();
        let actions_len = action_by_id.len();
        info!("The User Journey contains {} action(s).", &actions_len);
        
        if 0 < actions_len
        {
            let mut start_actions = Vec::new();
            for action in action_by_id.values()
            {
                // if a action has incoming arrows only from non-action objects, it is a start action
                if let Some(edges) = edges_by_target.get(&action.get_id())
                {
                    let mut is_start_action = true;
                    for edge in edges
                    {
                        if let Some(source) = edge.get_source()
                        {
                            if action_by_id.contains_key(&source)
                            {
                                is_start_action = false;
                            }
                        }
                    }
                    if is_start_action
                    {
                        start_actions.push(action);
                    }
                } else {
                // if a action has no incoming arrows, it is a start action
                    start_actions.push(action);
                }
            }

            return match start_actions.len()
            {
                0 => Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey with actions must have one start action.".to_string())),
                1 => 
                {
                    let edges_by_source = mx_graph_model.get_edge_mx_cells_ref_by_source_map();
                    let touchpoints_by_parent_action = mx_graph_model.get_touchpoints_by_parent_action_map();

                    let start_action = start_actions.get(0).unwrap();
                    add_action(start_action, user_journey, edges_by_source, action_by_id,
                        touchpoints_by_parent_action, mx_graph_model)
                }
                _ => Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey can only have one start action.".to_string()))
            };
        }
        // check touchpoints*/

        Ok(user_journey)
    }
}

fn add_phase(mut index: u32, drawio_phase: &drawio_xml::datatypes::Phase,
    user_journey: model::UserJourney, 
    edges_by_source: HashMap<String, Vec<&drawio_xml::datatypes::MxCell>>,
    phases_by_id: HashMap<String, &drawio_xml::datatypes::Phase>,
    actions_by_parent: HashMap<String, Vec<&drawio_xml::datatypes::Action>>,
    descriptions_by_action: HashMap<String, &drawio_xml::datatypes::Description>,
    touchpoints_by_parent_action: HashMap<String, Vec<&drawio_xml::datatypes::Touchpoint>>,
    mx_graph_model: &drawio_xml::MxGraphModel) -> Result<model::UserJourney, UseCaseExecError> 
{
    let mut user_journey = user_journey;
    if user_journey.contains_phase(&drawio_phase.get_id())
    {
        Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey must not contain circular connected phases.".to_string()))
    } else {
        let id = drawio_phase.get_id();

        //let phase_title = drawio_phase.get_obj_title();

        let mut locale_values = Vec::new();
        for ujlang in drawio_phase.get_ujlang_list() {
            if ujlang.get_value() != "" || ujlang.get_language() == "en" {
                locale_values.push(model::LocaleValue::new(ujlang.get_language(), ujlang.get_value()))
            }
        }

        let mut phase = model::Phase::new(id.clone(), index, locale_values, Vec::new());
        
        add_actions_to_phase(&mut phase, &actions_by_parent, &descriptions_by_action, &touchpoints_by_parent_action);
        user_journey.add_phase(phase);

        // handle next phases
        let mut next_phases = Vec::new();
        if let Some(edges) = edges_by_source.get(&id)
        {
            for edge in edges
            {
                if let Some (target) = edge.get_target()
                {
                    if let Some (phase) = phases_by_id.get(&target)
                    {
                        next_phases.push(phase);
                    }
                }
            }
        }
        match next_phases.len()
        {
            0 => Ok(user_journey),
            1 => 
            {
                let next_phase = next_phases.get(0).unwrap();
                index += 1;
                add_phase(index, next_phase, user_journey, edges_by_source, phases_by_id,
                    actions_by_parent, descriptions_by_action, touchpoints_by_parent_action, mx_graph_model)
            }
            _ => Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey can only have phases with one successor.".to_string()))
        }
    }
}

/*fn add_action(index: u32,
    drawio_action: &drawio_xml::datatypes::Action,
    user_journey: model::UserJourney, 
    edges_by_source: HashMap<String, Vec<&drawio_xml::datatypes::MxCell>>,
    actions_by_id: HashMap<String, &drawio_xml::datatypes::Action>,
    touchpoints_by_parent_action: HashMap<String, Vec<&drawio_xml::datatypes::Touchpoint>>,
    mx_graph_model: &drawio_xml::MxGraphModel) -> Result<model::UserJourney, UseCaseExecError> 
{
    let mut user_journey = user_journey;

    if user_journey.contains_action(&drawio_action.get_id())
    {
        Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey must not contain circular connected actions.".to_string()))
    } else {
        let id = drawio_action.get_id();

        let mut touchpoints: Vec<model::Touchpoint> = Vec::new();
        if let Some (drawio_touchpoints) = touchpoints_by_parent_action.get(&drawio_action.get_id())
        {
            info!("Action {} contains {} touchpoint(s).", &drawio_action.get_id(), &drawio_touchpoints.len());
            for drawio_touchpoint in drawio_touchpoints 
            {
                let touchpoint = model::Touchpoint::new(drawio_touchpoint.get_id(),
                    drawio_touchpoint.get_service_id(), drawio_touchpoint.get_link(), 
                    drawio_touchpoint.get_service_title(), drawio_touchpoint.get_service_description());
                touchpoints.push(touchpoint);
            }
        }
        let action = model::Action::new_optional(id.clone(), index, drawio_action.get_obj_title(),
            drawio_action.get_obj_description(), touchpoints);
        user_journey.add_action(action);

        // handle next actions
        let mut next_actions = Vec::new();
        if let Some(edges) = edges_by_source.get(&id)
        {
            for edge in edges
            {
                if let Some (target) = edge.get_target()
                {
                    if let Some (action) = actions_by_id.get(&target)
                    {
                        next_actions.push(action);
                    }
                }
            }
        }
        match next_actions.len()
        {
            0 => Ok(user_journey),
            1 => 
            {
                let next_action = next_actions.get(0).unwrap();
                add_action(next_action, user_journey, edges_by_source, actions_by_id,
                    touchpoints_by_parent_action, mx_graph_model)
            }
            _ => Err(UseCaseExecError::MxGraphModelInvalid("A valid User Journey can only have actions with one successor.".to_string()))
        }
    }
}*/

fn add_actions_to_phase(phase: &mut model::Phase,
    actions_by_parent: &HashMap<String, Vec<&drawio_xml::datatypes::Action>>,
    descriptions_by_action: &HashMap<String, &drawio_xml::datatypes::Description>,
    touchpoints_by_parent_action: &HashMap<String, Vec<&drawio_xml::datatypes::Touchpoint>>)
{
    if let Some(actions) = actions_by_parent.get(&phase.get_drawio_id())
    {
        info!("Phase {} contains {} action(s).", &phase.get_drawio_id(), &actions.len());

        let mut index = 0;
        for drawio_action in actions
        {
            let mut touchpoints: Vec<model::Touchpoint> = Vec::new();
            //info!("T by P: {:?}", touchpoints_by_parent_action);

            if let Some (drawio_touchpoints) = touchpoints_by_parent_action.get(&drawio_action.get_id())
            {
                info!("Action {} contains {} touchpoint(s).", &drawio_action.get_id(), &drawio_touchpoints.len());
                for drawio_touchpoint in drawio_touchpoints 
                {
                    let touchpoint = model::Touchpoint::new(drawio_touchpoint.get_id(),
                        drawio_touchpoint.get_service_id(), drawio_touchpoint.get_link(), drawio_touchpoint.get_location(),
                        drawio_touchpoint.get_service_title(), drawio_touchpoint.get_service_description());
                    touchpoints.push(touchpoint);
                }
            }

            let mut name_locale_values = Vec::new();
            for ujlang in drawio_action.get_ujlang_list() {
                if ujlang.get_value() != "" || ujlang.get_language() == "en" {
                    name_locale_values.push(model::LocaleValue::new(ujlang.get_language(), ujlang.get_value()))
                } 
            }

            let mut desc_locale_values = Vec::new();
            if let Some(description) = descriptions_by_action.get(&drawio_action.get_id()){
                for ujlang in description.get_ujlang_list() {
                    if ujlang.get_value() != "" || ujlang.get_language() == "en" {
                        desc_locale_values.push(model::LocaleValue::new(ujlang.get_language(), ujlang.get_value()))
                    } 
                }
            }

            if drawio_action.is_mandatory() {
                let action = model::Action::new_mandatory(drawio_action.get_id(),
                    index, name_locale_values, desc_locale_values, touchpoints);
                    phase.add_action(action);
            } else {
                let action = model::Action::new_optional(drawio_action.get_id(),
                    index, name_locale_values, desc_locale_values, touchpoints);
                    phase.add_action(action);
            }
            index += 1;
        }
    }
}