use crate::entities::bpmn::YaSerialize;
use crate::entities::bpmn::datatypes::events::start_event::StartEvent;
use crate::entities::bpmn::datatypes::events::end_event::EndEvent;
use crate::entities::bpmn::datatypes::tasks::{service_task::ServiceTask, sub_process::SubProcess};
use crate::entities::bpmn::datatypes::flows::sequence_flow::SequenceFlow;
use crate::entities::bpmn::config;

#[derive(Clone, Debug, YaSerialize, PartialEq)]
pub struct Process {
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(attribute)]
    pub name: Option<String>,
    #[yaserde(attribute, rename = "isExecutable")]
    pub is_executable: Option<String>,
    #[yaserde(attribute, rename = "xmlns:tns")]
    pub xmlns_tns: Option<String>,
    #[yaserde(attribute, rename = "tns:packageName")]
    pub tns_package_name: Option<String>,
    #[yaserde(attribute, rename = "processType")]
    pub process_type: Option<String>,
    /*
     * Assumption: ACROSS Workflows have one start- and one 
     * end-element
     */
    #[yaserde(child, rename = "bpmn2:startEvent")]
    pub start_event: Option<StartEvent>,
    #[yaserde(child, rename = "bpmn2:endEvent")]
    pub end_event: Option<EndEvent>,
    /* Each action will be represented as a task */
    #[yaserde(child, rename = "bpmn2:serviceTask")]
    pub tasks: Vec<ServiceTask>,
    #[yaserde(child, rename = "bpmn2:sequenceFlow")]
    pub sequence_flows: Vec<SequenceFlow>,
    /* Each phase will be represented as a sub process */
    #[yaserde(child, rename = "bpmn2:subProcess")]
    pub sub_processes: Vec<SubProcess>
}

impl Process {

    pub fn init(process_id: String, process_name: Option<String>) -> Process {
        Process {
            id:                 process_id,
            name:               process_name,
            is_executable:      Some(config::IS_EXECUTABLE.to_string()),
            xmlns_tns:          Some(config::XMLNS_TNS.to_string()),
            tns_package_name:   Some(config::TNS_PACKAGE_NAME.to_string()),
            process_type:       Some(config::PROCESS_TYPE.to_string()),
            start_event:        None,
            end_event:          None,
            tasks:              Vec::new(),
            sequence_flows:     Vec::new(),
            sub_processes:      Vec::new()
        }
    }

    pub fn get_tasks(self) -> Vec<ServiceTask> {
        self.tasks
    }

    pub fn push_to_tasks(&mut self, service_task: ServiceTask) {
        self.tasks.push(service_task);
    }

    pub fn push_to_sub_processes(&mut self, sub_process: SubProcess) {
        self.sub_processes.push(sub_process);
    }

    pub fn get_sequence_flows(self) -> Vec<SequenceFlow> {
        self.sequence_flows
    }

    pub fn push_to_flows(&mut self, sequence_flow: SequenceFlow) {
        self.sequence_flows.push(sequence_flow);
    }

}