use failure::Fail;
use yaserde_derive::YaDeserialize;
use yaserde;
use std::collections::HashMap;

pub mod datatypes;

#[derive(Fail, Debug)]
pub enum DrawioXMLError {
    #[fail(display = "The Drawio XML cannot be serialized. ({})", _0)]
    DeserializationFailed(String)
}

#[derive(Debug, YaDeserialize)]
#[yaserde(rename = "mxGraphModel")]
pub struct MxGraphModel {
    #[yaserde(child)]
    pub root: Option<Root>
}

impl MxGraphModel {

    pub fn get_phases(&self) -> Option<&Vec<datatypes::Phase>> {
        if let Some(root) = &self.root {
            return Some(&root.phases)
        } else {
            None
        }
    }

    pub fn get_user_journey_properties(&self) -> Option<&datatypes::UJProperties> {
        if let Some(root) = &self.root {
            if let Some(user_journey_properties) = &root.user_journey_properties {
                return Some(user_journey_properties)
            } else {
                None
            }
        } else {
            None
        }
    }

    pub fn get_phase_by_id_map(&self) -> HashMap<String, &datatypes::Phase> {
        let mut phases_by_id = HashMap::new();
        if let Some (phases) = self.get_phases() {
            for phase in phases {
                phases_by_id.insert(phase.get_id(), phase);
            }
        }
        return phases_by_id;
    }

    pub fn get_action_by_id_map(&self) -> HashMap<String, &datatypes::Action> {
        let mut action_by_id = HashMap::new();
        if let Some (actions) = self.get_actions() {
            for action in actions {
                action_by_id.insert(action.get_id(), action);
            }
        }
        return action_by_id;
    }

    pub fn get_description_by_action_id_map(&self) -> HashMap<String, &datatypes::Description> {
        let mut description_by_action_id = HashMap::new();
        let action_by_id_map = self.get_actions_by_id_map(); // checks if action exists
        if let Some (descriptions) = self.get_descriptions() {
            for description in descriptions {
                if let Some(parent_id) = description.mx_cell.get_parent() {
                    if let Some(action) = action_by_id_map.get(&parent_id) {
                        description_by_action_id.insert(action.get_id(), description);
                    }
                }
            }
        }
        return description_by_action_id;
    }

    pub fn get_mx_cells(&self) -> Option<&Vec<datatypes::MxCell>> {
        if let Some(root) = &self.root {
            return Some(&root.mx_cells)
        } else {
            None
        }
    }

    pub fn get_edge_mx_cells_ref(&self) -> Vec<&datatypes::MxCell> {
        let mx_cells = self.get_mx_cells();
        let mut result = Vec::new();
        if mx_cells.is_some() {
            for mx_cell in mx_cells.unwrap() {
                if mx_cell.is_edge() {
                    result.push(mx_cell);
                }
            }
        }
        return result;
    }

    pub fn get_edge_mx_cells_ref_by_target_map(&self) -> HashMap<String, Vec<&datatypes::MxCell>> {
        let mut edges_by_target: HashMap<String, Vec<&datatypes::MxCell>> = HashMap::new();
        for curr_edge in self.get_edge_mx_cells_ref() {
            if let Some(target) = &curr_edge.target {

                if let Some(edge) = edges_by_target.get_mut(target){
                    edge.push(curr_edge);
                } else {
                    edges_by_target.insert(target.to_string(), vec![curr_edge]);
                }
            }
        }
        return edges_by_target;
    }

    pub fn get_edge_mx_cells_ref_by_source_map(&self) -> HashMap<String, Vec<&datatypes::MxCell>> {
        let mut edges_by_source: HashMap<String, Vec<&datatypes::MxCell>> = HashMap::new();
        for curr_edge in self.get_edge_mx_cells_ref() {
            if let Some(source) = &curr_edge.source {
                if let Some(edge) = edges_by_source.get_mut(source) {
                    edge.push(curr_edge);
                } else {
                    edges_by_source.insert(source.to_string(), vec![curr_edge]);
                }
            }
        }
        return edges_by_source;
    }

    pub fn get_actions(&self) -> Option<&Vec<datatypes::Action>> {
        if let Some(root) = &self.root {
            return Some(&root.actions)
        } else {
            None
        }
    }

    pub fn get_descriptions(&self) -> Option<&Vec<datatypes::Description>> {
        if let Some(root) = &self.root {
            return Some(&root.descriptions)
        } else {
            None
        }
    }

    pub fn get_actions_by_id_map(&self) -> HashMap<String, &datatypes::Action> {
        let mut actions_by_id = HashMap::new();
        if let Some (actions) = self.get_actions() {
            for action in actions {
                actions_by_id.insert(action.get_id(), action);
            }
        }
        return actions_by_id;
    }

    pub fn get_actions_by_parent_map(&self) -> HashMap<String, Vec<&datatypes::Action>> {
        let mut actions_by_parent: HashMap<String, Vec<&datatypes::Action>> = HashMap::new();
        if let Some (actions) = self.get_actions() {
            for action in actions {
                if let Some(parent) = action.get_parent() {
                    if actions_by_parent.contains_key(&parent) {
                        let mut actions_vec = actions_by_parent.get(&parent).unwrap().to_vec();
                        actions_vec.push(action);
                        actions_by_parent.insert(parent, actions_vec);
                    } else {
                        actions_by_parent.insert(parent, vec![action]);
                    }
                }
            }
        }
        return actions_by_parent;
    }

    pub fn get_touchpoints(&self) -> Option<&Vec<datatypes::Touchpoint>> {
        if let Some(root) = &self.root {
            return Some(&root.touchpoints)
        } else {
            None
        }
    }

    pub fn get_touchpoints_by_parent_action_map(&self) -> HashMap<String, Vec<&datatypes::Touchpoint>>
    {
        let mut touchpoints_by_parent_action: HashMap<String, Vec<&datatypes::Touchpoint>> = HashMap::new();
        
        if let Some (touchpoint_lists) = self.get_touchpoint_lists() {
            info!("t-lists: {:?}", touchpoint_lists);
            if let Some (touchpoints) = self.get_touchpoints() {
                info!("ts: {:?}", touchpoints);
                let actions_by_id = self.get_actions_by_id_map();
                info!("as by ids: {:?}", actions_by_id);
                for curr_touchpoint in touchpoints {
                    // get parent touchpoint list
                    if let Some(touchpoint_list_id) = curr_touchpoint.get_parent() {
                        for touchpoint_list in touchpoint_lists {
                            if touchpoint_list.get_id().eq(&touchpoint_list_id) {
                                // get parent action id
                                if let Some(action_id) = touchpoint_list.get_parent() {
                                    // check if action
                                    if actions_by_id.contains_key(&action_id) {
                                        if let Some(touchpoints) = touchpoints_by_parent_action.get_mut(&action_id) {
                                            touchpoints.push(curr_touchpoint);
                                        } else {
                                            touchpoints_by_parent_action.insert(action_id, vec![curr_touchpoint]);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return touchpoints_by_parent_action;
    }

    pub fn get_touchpoint_lists(&self) -> Option<&Vec<datatypes::TouchpointList>> {
        if let Some(root) = &self.root {
            return Some(&root.touchpoint_lists)
        } else {
            None
        }
    }

}

#[derive(Debug, YaDeserialize)]
#[yaserde(rename = "root")]
pub struct Root {
    #[yaserde(child, rename="mxCell")]
    pub mx_cells: Vec<datatypes::MxCell>,
    #[yaserde(child, rename="object")]
    pub objects: Vec<datatypes::Object>,
    #[yaserde(child, rename="phase")]
    pub phases: Vec<datatypes::Phase>,
    #[yaserde(child, rename="action")]
    pub actions: Vec<datatypes::Action>,
    #[yaserde(child, rename="description")]
    pub descriptions: Vec<datatypes::Description>,
    #[yaserde(child, rename="userinfo")]
    pub user_infos: Vec<datatypes::UserInfo>,
    #[yaserde(child, rename="touchpoints")]
    pub touchpoint_lists: Vec<datatypes::TouchpointList>,
    #[yaserde(child, rename="ServiceObject")]
    pub touchpoints: Vec<datatypes::Touchpoint>,
    #[yaserde(child, rename="ujproperties")]
    pub user_journey_properties: Option<datatypes::UJProperties>,
}

pub fn deserialize(input: String) -> Result<MxGraphModel, DrawioXMLError> {
    let mx_graph_model: MxGraphModel = yaserde::de::from_reader(input.as_bytes())
        .map_err(|error| DrawioXMLError::DeserializationFailed(error))?;
    Ok(mx_graph_model)
}