use failure::Fail;
use serde_json;
use serde_json::Value;
use serde_derive::Serialize;

use crate::entities::drawio_xml::datatypes::{Object, Obejct};

#[derive(Fail, Debug)]
pub enum WorkflowError {
    #[fail(display = "The JSON Workflow cannot be deserialized. ({})", _0)]
    DeserializationFailed(serde_json::error::Error),
    #[fail(display = "The JSON Workflow cannot be serialized. ({})", _0)]
    SerializationFailed(serde_json::error::Error)
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Workflow {
    pub origin_country: String,
    pub destination_country: String,
    pub workflow_type: String,
    #[serde(rename = "$value")]
    pub actions: Vec<Action>
}

impl Workflow {
    pub fn new() -> Workflow {
        Workflow {
            origin_country: "".to_string(),
            destination_country: "".to_string(),
            workflow_type: "".to_string(),
            actions: Vec::new()
        }
    }

    pub fn get_origin_country(&self) -> String {
        self.origin_country.to_string()
    }

    pub fn get_destination_country(&self) -> String {
        self.destination_country.to_string()
    }

    pub fn get_workflow_type(&self) -> String {
        self.workflow_type.to_string()
    }

    pub fn add_action(&mut self, action: Action) {
        self.actions.push(action)
    }
}

#[derive(Debug, Serialize, Deserialize, PartialEq)]
pub struct Action {
    pub id: String,
    pub name: String,
    pub service_ids: Vec<String>
}

impl Action {

    pub fn new(id: String, name: String, service_ids: Vec<String>) -> Action {
        Action {
            id: id,
            name: name,
            service_ids: service_ids
        }
    }

    pub fn new_initial(id: String, name: String) -> Action {
        Action {
            id: id,
            name: name,
            service_ids: Vec::new()
        }
    }

    pub fn add_service_id(&mut self, service_id: String) {
        self.service_ids.push(service_id)
    }

    pub fn new_from_object(object: Object) -> Option<Action> {
        if object.service_id.is_some() && object.mx_cell.value.is_some() {
            let service_id = object.service_id.unwrap();
            let mut action = Action::new_initial(service_id.clone(), object.mx_cell.value.unwrap());
            action.add_service_id(service_id);
            Some(action)
        } else {
            return None;
        }
    }

    pub fn new_from_obejct(obejct: Obejct) -> Option<Action> {
        if obejct.service_id.is_some() && obejct.mx_cell.value.is_some() {
            let service_id = obejct.service_id.unwrap();
            let mut action = Action::new_initial(service_id.clone(), obejct.mx_cell.value.unwrap());
            action.add_service_id(service_id);
            Some(action)
        } else {
            return None;
        }
    }

    pub fn get_name(&self) -> String {
        self.name.to_string()
    }

    pub fn get_service_ids(&self) -> Vec<String> {
        self.service_ids.to_vec()
    }
}

pub fn deserialize(workflow_json: String) -> Result<Workflow, WorkflowError>
{
    let mut workflow = Workflow::new();
    let workflow_json_value = serde_json::from_str(&workflow_json)
        .map_err(|err| WorkflowError::DeserializationFailed(err))?;
    process(workflow_json_value, &mut workflow);
    Ok(workflow)
}

pub fn process(value: Value, workflow: &mut Workflow)
{
    match value {
        Value::Null             => return,
        Value::Bool(_x)         => return,
        Value::Number(_x)       => return,
        Value::String(_x)       => return,
        Value::Array(array)     => process_array(array, workflow),
        Value::Object(map)      => process_map(map, workflow),
    }
}

pub fn process_map(map: serde_json::Map<std::string::String, serde_json::Value>, workflow: &mut Workflow)
{
    if let Some(id) = map.get("id") {
        if let Some(name_value) = map.get("name") {
            if let Some(service_id_value) = map.get("service_id") {
                if let Some(id_str) = id.as_str() {
                    if let Some(name_str) = name_value.as_str() {
                        if let Some(service_id_array) = service_id_value.as_array() {
                            let mut action = Action::new_initial(id_str.to_string(), name_str.to_string());
                            for service_id in service_id_array {
                                action.add_service_id(service_id.to_string());
                            }
                            workflow.actions.push(action);
                        }
                    }
                }
            }
        }
    }
}

fn process_array(array: Vec<Value>, workflow: &mut Workflow)
{
    for value in array {
        process(value, workflow)
    }
}