extern crate serde;
extern crate serde_json;
extern crate yaserde;

use crate::cli::UJMTError;
use crate::core::config;
//use crate::core::log;
use crate::domain::usecases;
use crate::entities::messages;

use chrono::Utc;
use clap::{ArgAction, ArgMatches, Command};
use std::convert::Infallible;
use std::net::SocketAddr;
use warp::{http::StatusCode, Filter};

use uuid::Uuid;

use tokio_postgres::NoTls;

// cli defintion
pub fn init() -> clap::Command {
    Command::new("serve")
        .about("Logging received message bodies and request information")
        .arg(
            clap::Arg::new("log file")
                .short('l')
                .action(ArgAction::Set)
                .help("sets the name of an optional log file to use instead of stdout"),
        )
        .arg(
            clap::Arg::new("port")
                .short('p')
                .long("port")
                .action(ArgAction::Set)
                .default_value(config::DEFAULT_PORT)
                .help("sets the port number"),
        )
}

//----------------------------------------------------------
// Web logs
fn log_request() -> impl Filter<Extract = (String,), Error = Infallible> + Copy {
    warp::method().and(warp::path::full()).map(
        |method: http::method::Method, full: warp::filters::path::FullPath| {
            info!("Received Request"; "method" => method.as_str(), "path" => full.as_str());
            format!("{:?}", full)
        },
    )
}

fn log_not_found() -> impl Filter<Extract = (String,), Error = Infallible> + Copy {
    warp::method().and(warp::path::full()).map(
        |method: http::method::Method, full: warp::filters::path::FullPath| {
            error!("Request Not Found"; "method" => method.as_str(), "path" => full.as_str());
            format!("{:?}", full)
        },
    )
}

// serve command
#[tokio::main]
pub async fn exec(matches: &ArgMatches) -> Result<(), UJMTError> {
    info!("Starting {}", config::SERVICE);

    // Get environment
    let port = matches.get_one::<String>("port").unwrap();
    let _service_name = config::SERVICE;

    // Init logging
    let mut _log_level = "info";
    if matches.get_flag("debugging") {
        _log_level = "debug";
    };

    /*
    let mut _guard;
    if let Some(file) = matches.value_of("log file") {
        _guard =
            slog_scope::set_global_logger(log::new_file(log_level, file));
    } else {
        _guard = slog_scope::set_global_logger(log::new_term(log_level));
    }
    debug!("Initialized Global Logger");
    */

    //------------------------------
    // Set routing for web server

    let root = warp::path::end() // Matches '/'
        .and(warp::get())
        .and(log_request())
        .and_then(root_handler);

    let root_option = warp::path::end() // Matches '/'
        .and(warp::options())
        .map(|| warp::reply())
        .map(|reply| 
            warp::reply::with_header(reply, "Allow", "OPTIONS, GET, POST, PUT, DELETE"))
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Methods", "*")
        })
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Origin", config::FRONTEND_URL)
        })
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Headers", "Content-Type")
        })
        .map(|reply| 
            warp::reply::with_status(reply, StatusCode::NO_CONTENT)
        );

    let not_found = warp::any().and(log_not_found()).and_then(not_found_handler);

    //------------------------------
    // For Developers

    // Endpoint for developers who test the BPMN generation from DRAWIO XML.
    let generate = warp::path("generate") // Matches '/generate'
        .and(warp::post())
        .and(warp::body::content_length_limit(1024 * 32))
        .and(log_request())
        .and(warp::body::bytes()) // DRAWIO XML
        .and_then(generate_handler);

    //------------------------------
    // For Modellers

    // Endpoint for Modellers / UJMT Frontend to mark the Workflow in a DRAWIO XML.
    let mark_workflow = warp::path!("validate") // Matches '/validate'
        .and(warp::post())
        //.and(warp::body::content_length_limit(1024 * 32))
        .and(log_request())
        .and(warp::body::bytes()) // DRAWIO XML
        .and_then(mark_handler);

    let mark_workflow_option = warp::path!("validate") // Matches '/validate'
        .and(warp::options())
        //.and(log_request())
        .map(|| warp::reply())
        .map(|reply| warp::reply::with_header(reply, "Allow", "OPTIONS, POST"))
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Origin", config::FRONTEND_URL)
        })
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Headers", "Content-Type")
        })
        .map(|reply| warp::reply::with_status(reply, StatusCode::NO_CONTENT));

    // Engine Interface

    // Endpoint for Modellers / UJMT Frontend to get all stored BPMN files.
    let get_all_workflows = warp::path("workflows") // Matches '/workflows'
        .and(warp::get())
        .and(log_request())
        .and_then(get_all_workflows_handler);

    // Endpoint for Modellers / UJMT Frontend to submit Workflows to the Engine
    let create_workflow = warp::path!("workflows" / String) // Matches '/workflows'
        .and(warp::post())
        //.and(warp::body::content_length_limit(1024 * 64)) //64 kb
        //.and(warp::query::<std::collections::HashMap<String, String>>())
        .and(log_request())
        .and(warp::body::bytes()) // JSON
        .and_then(create_workflow_handler);

    // Endpoint for Modellers / UJMT Frontend to delete Workflows in the Engine
    let delete_workflow = warp::path!("workflows" / String) // Matches '/workflows/{id}'
        .and(warp::delete())
        .and(log_request())
        .and(warp::body::bytes()) // JSON
        .and_then(delete_workflow_handler);

    // Storage Interface

    // Endpoint for Modellers / UJMT Frontend to load all User Journeys
    let get_all_user_journeys = warp::path!("userjourneys") // Matches '/userjourneys'
        .and(warp::get())
        .and(log_request())
        .and_then(get_all_user_journeys_handler);

    let init_storage = warp::path!("userjourneys" / "reset") // Matches '/userjourneys/reset'
        //.and(warp::path::end())
        .and(warp::post())
        .and(log_request())
        .and_then(init_storage_handler);

    // Endpoint for Modellers / UJMT Frontend to store User Journeys
    let store_user_journey = warp::path!("userjourneys") // Matches '/userjourneys'
        .and(warp::post())
        //.and(warp::body::content_length_limit(1024 * 32))
        .and(log_request())
        .and(warp::body::bytes()) // DRAWIO XML
        .and_then(store_user_journey_handler);

    // Endpoint for Modellers / UJMT Frontend to store User Journeys
    let update_user_journey = warp::path!("userjourneys" / String) // Matches '/userjourneys/{id}'
        .and(warp::put())
        //.and(warp::body::content_length_limit(1024 * 32))
        .and(log_request())
        .and(warp::body::bytes()) // DRAWIO XML
        .and_then(update_user_journey_handler);

    let user_journey_id_option = warp::path!("userjourneys" / String) // /userjourneys/{id}'
        .and(warp::options())
        //.and(log_request())
        .map(|_| warp::reply())
        .map(|reply| 
            warp::reply::with_header(reply, "Allow", "OPTIONS, GET, POST, PUT, DELETE"))
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Methods", "*")
        })
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Origin", config::FRONTEND_URL)
        })
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Headers", "Content-Type")
        })
        .map(|reply| 
            warp::reply::with_status(reply, StatusCode::NO_CONTENT)
        );

    let user_journey_option = warp::path("userjourneys") // Matches '/userjourneys'
        .and(warp::options())
        //.and(log_request())
        .map(|| warp::reply())
        .map(|reply| 
            warp::reply::with_header(reply, "Allow", "OPTIONS, GET, POST, PUT, DELETE"))
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Methods", "*")
        })
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Origin", config::FRONTEND_URL)
        })
        .map(|reply| {
            warp::reply::with_header(reply, "Access-Control-Allow-Headers", "Content-Type")
        })
        .map(|reply| 
            warp::reply::with_status(reply, StatusCode::NO_CONTENT)
        );

    // Endpoint for Modellers / UJMT Frontend to load a specific User Journey
    let get_user_journey = warp::path!("userjourneys" / String) // Matches '/userjourneys/{id}'
        .and(warp::get())
        .and(warp::query::<std::collections::HashMap<String, String>>())
        .and(log_request())
        .and_then(get_user_journey_handler);

    // Endpoint for Modellers / UJMT Frontend to delete a specific User Journey
    let delete_user_journey = warp::path!("userjourneys" / String) // Matches '/userjourneys/{id}'
        .and(warp::delete())
        .and(log_request())
        .and(warp::body::bytes()) // JSON
        .and_then(delete_user_journey_handler);

    // Start web server
    info!("Starting web server"; "port" => port);
    let addr: SocketAddr = format!("0.0.0.0:{}", port).parse().unwrap();
    warp::serve(
        root.or(root_option)
            .or(generate)
            .or(init_storage)
            .or(mark_workflow)
            .or(mark_workflow_option)
            .or(create_workflow)
            .or(get_all_workflows)
            .or(delete_workflow)
            .or(store_user_journey)
            .or(user_journey_id_option)
            .or(user_journey_option)
            .or(get_all_user_journeys)
            .or(update_user_journey)
            .or(get_user_journey)
            .or(delete_user_journey)
            .or(not_found),
    ) // not_found must be last
    .run(addr)
    .await;

    Ok(())
}

//----------------------------------------------------------
// Web Handlers

pub async fn root_handler(_tail: String) -> Result<impl warp::Reply, Infallible> {
    info!("Log Root Get Request");
    let json = warp::reply::json(&messages::StatusMessage::ok());
    Ok(warp::reply::with_status(json, StatusCode::OK))
}

pub async fn not_found_handler(_tail: String) -> Result<impl warp::Reply, Infallible> {
    let json = warp::reply::json(&messages::StatusMessage::not_found());
    Ok(warp::reply::with_status(json, StatusCode::NOT_FOUND))
}

pub async fn generate_handler(
    _tail: String,
    body: warp::hyper::body::Bytes,
) -> Result<impl warp::Reply, Infallible> {
    let bad_request_message = warp::reply::json(&messages::StatusMessage::bad_request());

    // read body as string from bytes
    let body_string = String::from_utf8(body.to_vec());
    if body_string.is_err() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_string = body_string.unwrap_or_default();
    info!("Log Request with"; "body" => body_string.clone());

    // Generate Workflow from Drawio XML or JSON
    // let user_journey = usecases::generate_user_journey_from_drawio(body_string.clone());
    let user_journey = usecases::generate_user_journey_from_drawio(body_string.clone());
    if user_journey.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    //let user_journey = user_journey.unwrap();
    let user_journey = user_journey.unwrap();

    // Return UJSEMessage as JSON
    info!("Log Result Ok");
    let message = warp::reply::json(&user_journey);
    return Ok(warp::reply::with_status(message, StatusCode::OK));
}

// Receives a Drawio XML input and generates and returns the JSON Workflow
pub async fn mark_handler(
    _tail: String,
    body: warp::hyper::body::Bytes,
) -> Result<
    warp::reply::WithStatus<warp::reply::WithHeader<warp::reply::WithHeader<warp::reply::Json>>>,
    Infallible,
> {
    let bad_request_message = warp::reply::with_header(
        warp::reply::with_header(
            warp::reply::json(&messages::StatusMessage::bad_request()),
            "Access-Control-Allow-Origin",
            "http://localhost:8080", // TODO: Parametrisieren
        ),
        "Access-Control-Allow-Headers",
        "Content-Type",
    );

    // read body as string from bytes
    let body_string = String::from_utf8(body.to_vec());
    if body_string.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_string = body_string.unwrap();
    info!("Log Request with"; "body" => body_string.clone());

    let mark_response = usecases::generate_mark_response_from_drawio(body_string.clone());
    if mark_response.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let mark_response = mark_response.unwrap();

    info!("Log Result Ok");

    let message = warp::reply::with_header(
        warp::reply::with_header(
            warp::reply::json(&mark_response),
            "Access-Control-Allow-Origin",
            "http://localhost:8080",
        ),
        "Access-Control-Allow-Headers",
        "Content-Type",
    );

    return Ok(warp::reply::with_status(message, StatusCode::OK));
}

/* Reads a Drawio XML from the storage, generates and returns the JSON and sends the output to the Engine */
pub async fn create_workflow_handler(
    user_journey_id: String,
    _tail: String,
    body: warp::hyper::body::Bytes,
) -> Result<impl warp::Reply, Infallible> {
    info!("user_journey_id: {:?}", user_journey_id);

    // status responses
    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());
    let bad_request_message = warp::reply::json(&messages::StatusMessage::bad_request());
    let not_found_message = warp::reply::json(&messages::StatusMessage::not_found());
    let forbidden_message = warp::reply::json(&messages::StatusMessage::forbidden());

    // read json body as string from bytes
    let body_string = String::from_utf8(body.to_vec());
    if body_string.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_string = body_string.unwrap_or_default();
    info!("Log Request with"; "body" => body_string.clone());
    let body_json_res = serde_json::from_str(&body_string);
    if body_json_res.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_json: serde_json::Value = body_json_res.unwrap();
    info!("body_json: {:?}", body_json);

    let timestamp: Option<&str> = body_json["timestamp"].as_str();
    let origin_country: Option<&str> = body_json["origin_country"].as_str();
    let destination_country: Option<&str> = body_json["destination_country"].as_str();
    let workflow_type: Option<&str> = body_json["workflow_type"].as_str();

    let id: Result<Uuid, uuid::Error> = Uuid::parse_str(user_journey_id.as_str());
    //let id = user_journey_id.parse::<i64>();

    if timestamp.is_none()
        || origin_country.is_none()
        || id.is_err()
        || destination_country.is_none()
        || workflow_type.is_none()
    {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let id = id.unwrap();
    let timestamp = timestamp.unwrap().to_string();
    let origin_country = origin_country.unwrap().to_string();
    let destination_country = destination_country.unwrap().to_string();
    let workflow_type = workflow_type.unwrap().to_string();

    // parse url
    let url = reqwest::Url::parse(config::ENGINE_URL);
    if url.is_err() {
        info!("Log (Server Error)");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
    let url = url.unwrap();

    // connect to db
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(_e) = connection.await {
                info!("Log Connection Error");
            }
        });

        // check if publishable
        let rows = client
            .query(
                "SELECT publishable FROM publishable_models 
                WHERE workflow_type=$1 AND origin_country=$2 AND destination_country=$3;",
                &[&workflow_type, &origin_country, &destination_country],
            )
            .await;
        info!("rows: {:?}", rows);

        if let Ok(msg) = rows {
            info!("msg: {:?}", msg);
            if msg.len() > 1 {
                // unexpected result
                info!("Log Result Error (msg.len() > 1)");
                return Ok(warp::reply::with_status(
                    internal_server_error_message,
                    StatusCode::INTERNAL_SERVER_ERROR,
                ));
            } else {
                if let Some(row) = msg.get(0) {
                    let publishable: bool = row.get("publishable");
                    info!("publishable: {}", publishable);
                    if publishable {
                        let rows = client
                            .execute("UPDATE publishable_models
                                SET publishable=FALSE
                                WHERE origin_country=$1 AND destination_country=$2 AND workflow_type=$3;", 
                                &[&origin_country, &destination_country, &workflow_type],)
                            .await;

                        if let Ok(_msg) = rows {
                            let rows = client
                                .query("SELECT json_description, uj_type
                                    FROM user_journey_models
                                    WHERE timestamp=$1 AND id=$2;", &[&timestamp, &id])
                                .await;
                            if let Ok(msg) = rows {
                                if msg.len() > 1 {
                                    info!("Log Result Error");
                                    return Ok(warp::reply::with_status(
                                        internal_server_error_message,
                                        StatusCode::INTERNAL_SERVER_ERROR,
                                    ));
                                } else {
                                    if let Some(row) = msg.get(0) {
                                        info!("row: {:?}", row);

                                        let uj_type: String = row.get("uj_type");
                                        let json_description: Option<String> =
                                            row.get("json_description");

                                        if json_description.is_none() || uj_type == "abstract" {
                                            info!("Log Request Forbidden");
                                            let rows = client
                                                .execute("UPDATE publishable_models
                                                    SET publishable=TRUE
                                                    WHERE origin_country=$1 AND destination_country=$2 AND workflow_type=$3;", &[&origin_country, &destination_country, &workflow_type],)
                                                .await;
                                            if let Ok(_msg) = rows {
                                                info!("Log Request Forbidden");
                                                return Ok(warp::reply::with_status(
                                                    forbidden_message,
                                                    StatusCode::FORBIDDEN,
                                                ));
                                            } else {
                                                info!("Log Request Server Error");
                                                return Ok(warp::reply::with_status(
                                                    internal_server_error_message,
                                                    StatusCode::INTERNAL_SERVER_ERROR,
                                                ));
                                            }
                                        }
                                        let json_description = json_description.unwrap();

                                        // send to engine and get engine_id
                                        let part = reqwest::multipart::Part::text(json_description)
                                            .file_name(format!(
                                                "json_{}.json",
                                                Utc::now().format("%Y-%m-%d_%H-%M-%S")
                                            ));
                                        let form =
                                            reqwest::multipart::Form::new().part("json file", part);

                                        let req_client = reqwest::Client::new();
                                        let res = req_client.post(url).multipart(form).send().await;

                                        if res.is_err() {
                                            info!("Log Result Error");
                                            let err_status = res.unwrap_err().status();
                                            if err_status.is_none() {
                                                info!("Log (Server Error)");
                                                return Ok(warp::reply::with_status(
                                                    internal_server_error_message,
                                                    StatusCode::INTERNAL_SERVER_ERROR,
                                                ));
                                            } else {
                                                let err_status = err_status.unwrap();
                                                let status_message = warp::reply::json(
                                                    &messages::StatusMessage::new(err_status),
                                                );
                                                return Ok(warp::reply::with_status(
                                                    status_message,
                                                    err_status,
                                                ));
                                            }
                                        } else {
                                            let res = res.unwrap();
                                            let status = res.status();
                                            let engine_id = res.text().await;

                                            if engine_id.is_err() {
                                                info!("Log Engine ID Error");
                                                let err_status = engine_id.unwrap_err().status();
                                                if err_status.is_none() {
                                                    info!("Log (Server Error)");
                                                    return Ok(warp::reply::with_status(
                                                        internal_server_error_message,
                                                        StatusCode::INTERNAL_SERVER_ERROR,
                                                    ));
                                                } else {
                                                    let err_status = err_status.unwrap();
                                                    let status_message = warp::reply::json(
                                                        &messages::StatusMessage::new(err_status),
                                                    );
                                                    return Ok(warp::reply::with_status(
                                                        status_message,
                                                        err_status,
                                                    ));
                                                }
                                            } else {
                                                // update workflow ID in db
                                                let rows = client
                                                    .execute("UPDATE user_journey_models
                                                        SET engine_id=$1, status='published'
                                                        WHERE timestamp=$2 AND id=$3;",
                                                    &[&engine_id.unwrap(), &timestamp, &id],)
                                                    .await;

                                                if let Ok(msg) = rows {
                                                    info!("msg: {:?}", msg);
                                                    let status_message = warp::reply::json(
                                                        &messages::StatusMessage::new(status),
                                                    );
                                                    return Ok(warp::reply::with_status(
                                                        status_message,
                                                        status,
                                                    ));
                                                } else {
                                                    info!("Log Error Engine ID Update");
                                                    return Ok(warp::reply::with_status(
                                                        internal_server_error_message,
                                                        StatusCode::INTERNAL_SERVER_ERROR,
                                                    ));
                                                }
                                            }
                                        }
                                    } else {
                                        info!("Log Not Found");
                                        return Ok(warp::reply::with_status(
                                            not_found_message,
                                            StatusCode::NOT_FOUND,
                                        ));
                                    }
                                }
                            } else {
                                info!("Log Result Server Error");
                                return Ok(warp::reply::with_status(
                                    internal_server_error_message,
                                    StatusCode::INTERNAL_SERVER_ERROR,
                                ));
                            }
                        } else {
                            info!("Log Result Server Error");
                            return Ok(warp::reply::with_status(
                                internal_server_error_message,
                                StatusCode::INTERNAL_SERVER_ERROR,
                            ));
                        }
                    } else {
                        info!("Log Result Forbidden");
                        return Ok(warp::reply::with_status(
                            forbidden_message,
                            StatusCode::FORBIDDEN,
                        ));
                    }
                } else {
                    // not found
                    info!("Log Result Not Found");
                    return Ok(warp::reply::with_status(
                        not_found_message,
                        StatusCode::NOT_FOUND,
                    ));
                }
            }
        } else {
            info!("Log Result Error");
            return Ok(warp::reply::with_status(
                internal_server_error_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}

/* Reads a Drawio XML from the storage and deletes the associated workflow in the Engine */
pub async fn delete_workflow_handler(
    user_journey_id: String,
    _tail: String,
    body: warp::hyper::body::Bytes,
) -> Result<impl warp::Reply, Infallible> {
    info!("user_journey_id: {:?}", user_journey_id);

    let _ok_message = warp::reply::json(&messages::StatusMessage::ok());
    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());
    let bad_request_message = warp::reply::json(&messages::StatusMessage::bad_request());
    let not_found_message = warp::reply::json(&messages::StatusMessage::not_found());

    // parse id
    let id: Result<Uuid, uuid::Error> = Uuid::parse_str(user_journey_id.as_str());
    //let id = user_journey_id.parse::<i64>();
    if id.is_err() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let id = id.unwrap();
    info!("id: {:?}", id);

    // read json body as string from bytes
    let body_string = String::from_utf8(body.to_vec());
    if body_string.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_string = body_string.unwrap_or_default();
    info!("Log Request with"; "body" => body_string.clone());
    let body_json_res = serde_json::from_str(&body_string);
    if body_json_res.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_json: serde_json::Value = body_json_res.unwrap();
    info!("body_json: {:?}", body_json);

    let timestamp: Option<&str> = body_json["timestamp"].as_str();
    if timestamp.is_none() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let timestamp: String = timestamp.unwrap().to_string();
    info!("timestamp: {:?}", timestamp);

    // connect to db
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(_e) = connection.await {
                info!("Log Connection Error");
            }
        });

        // read user journey properties from storage
        let rows = client
            .query("SELECT engine_id, origin_country, destination_country, workflow_type
                FROM user_journey_models
                WHERE timestamp=$1 AND id=$2;", &[&timestamp, &id],).await;
        info!("rows: {:?}", rows);

        if let Ok(msg) = rows {
            info!("msg: {:?}", msg);
            if msg.len() > 1 {
                // unexpected result
                info!("Log Result Error (msg.len() > 1)");
                return Ok(warp::reply::with_status(
                    internal_server_error_message,
                    StatusCode::INTERNAL_SERVER_ERROR,
                ));
            } else {
                if let Some(row) = msg.get(0) {
                    info!("row: {:?}", row);

                    // get properties
                    let engine_id: String = row.get("engine_id");
                    let origin_country: String = row.get("origin_country");
                    let destination_country: String = row.get("destination_country");
                    let workflow_type: String = row.get("workflow_type");

                    info!("engine_id: {:?}", engine_id);
                    info!("origin_country: {:?}", origin_country);
                    info!("destination_country: {:?}", destination_country);
                    info!("workflow_type: {:?}", workflow_type);

                    // set uj status to 'deactivated' and engine_id to NULL in db
                    let rows = client
                        .execute("UPDATE user_journey_models 
                        SET status='deactivated', engine_id=NULL, trust_indicator=FALSE
                        WHERE timestamp=$1 AND id=$2;", &[&timestamp, &id],).await;
                    info!("rows: {:?}", rows);

                    if let Ok(msg) = rows {
                        info!("msg: {:?}", msg);

                        // create engine url
                        let url =
                            reqwest::Url::parse(&format!("{}/{}", config::ENGINE_URL, engine_id));
                        if url.is_err() {
                            info!("Log BAD REQUEST");
                            return Ok(warp::reply::with_status(
                                bad_request_message,
                                StatusCode::BAD_REQUEST,
                            ));
                        }

                        // delete workflow from engine
                        let engine_client = reqwest::Client::new();
                        let res = engine_client.delete(url.unwrap()).send().await;

                        if res.is_err() {
                            info!("Log Result Error");
                            let err_status = res.unwrap_err().status().unwrap_or_default();
                            let status_message =
                                warp::reply::json(&messages::StatusMessage::new(err_status));
                            return Ok(warp::reply::with_status(status_message, err_status));
                        } else {
                            // change all related uj 'publishable' to true
                            let rows = client
                                .execute("UPDATE publishable_models
                                    SET publishable=TRUE
                                    WHERE origin_country=$1 AND destination_country=$2 AND workflow_type=$3;", &[&origin_country, &destination_country, &workflow_type],).await;
                            info!("rows: {:?}", rows);

                            if let Ok(msg) = rows {
                                info!("Log Result Ok");
                                info!("msg: {:?}", msg);
                                let status = res.unwrap().status();
                                let status_message =
                                    warp::reply::json(&messages::StatusMessage::new(status));
                                return Ok(warp::reply::with_status(status_message, status));
                            } else {
                                info!("Log Result Error Status Update");
                                return Ok(warp::reply::with_status(
                                    internal_server_error_message,
                                    StatusCode::INTERNAL_SERVER_ERROR,
                                ));
                            }
                        }
                    } else {
                        info!("Log Result Error Status Update");
                        return Ok(warp::reply::with_status(
                            internal_server_error_message,
                            StatusCode::INTERNAL_SERVER_ERROR,
                        ));
                    }
                } else {
                    // not found
                    info!("Log Result Not Found");
                    return Ok(warp::reply::with_status(
                        not_found_message,
                        StatusCode::NOT_FOUND,
                    ));
                }
            }
        } else {
            info!("Log Result Error");
            return Ok(warp::reply::with_status(
                internal_server_error_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}

// for development/testing purposes only
pub async fn get_all_workflows_handler(_tail: String) -> Result<impl warp::Reply, Infallible> {
    info!("Log request");
    let res = reqwest::get(config::ENGINE_URL).await;

    if res.is_err() {
        info!("Log Result Error");
        let err_status = res.unwrap_err().status().unwrap_or_default();
        let status_message = warp::reply::json(&messages::StatusMessage::new(err_status));
        return Ok(warp::reply::with_status(status_message, err_status));
    } else {
        let res = res.unwrap();
        let status = res.status();
        let workflow = res.json::<Vec<messages::UJSEResponse>>().await;

        if workflow.is_err() {
            info!("Log Result Error");
            let status_message =
                warp::reply::json(&messages::StatusMessage::internal_server_error());
            return Ok(warp::reply::with_status(
                status_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        } else {
            info!("Log Result Ok");
            let workflow = workflow.unwrap();
            let workflow_message =
                warp::reply::json(&messages::AllWorkflowsUJSEResponse::new(status, workflow));
            return Ok(warp::reply::with_status(workflow_message, status));
        }
    }
}

// for development/testing purposes only
pub async fn init_storage_handler(_tail: String) -> Result<impl warp::Reply, Infallible> {
    info!("Log request");
    let ok_message = warp::reply::json(&messages::StatusMessage::ok());
    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());

    // connect to DB
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(e) = connection.await {
                info!("Log Result Error");
                eprintln!("connection error: {}", e);
            }
        });

        // drop existing table
        let rows = client.execute("DROP TABLE user_journey_models;", &[]).await;
        if let Err(_msg) = rows {
            info!("Drop Table Error");
        } else {
            info!("Drop Table Success");
        }

        // drop existing table
        let rows = client.execute("DROP TABLE publishable_models;", &[]).await;
        if let Err(_msg) = rows {
            info!("Drop Table Error");
        } else {
            info!("Drop Table Success");
        }

        let rows = client
            .execute(
                "CREATE TABLE user_journey_models (
                id                  UUID NOT NULL DEFAULT gen_random_uuid(),
                title               VARCHAR NOT NULL,
                drawio_diagram      TEXT NOT NULL,
                uj_type             VARCHAR NOT NULL,
                workflow_type       VARCHAR,
                origin_country      VARCHAR,
                destination_country VARCHAR,
                status              VARCHAR NOT NULL,
                trust_indicator     BOOLEAN,
                engine_id           VARCHAR,
                json_description    TEXT,
                timestamp           VARCHAR NOT NULL
            );",
                &[],
            )
            .await;
        info!("rows: {:?}", rows);

        if let Err(_msg) = rows {
            info!("Create Table Error");
        } else {
            info!("Create Table Success");
        }

        let rows = client
            .execute(
                "CREATE TABLE publishable_models (
                    origin_country      VARCHAR,
                    destination_country VARCHAR,
                    workflow_type       VARCHAR,
                    publishable         BOOLEAN NOT NULL
                );",
                &[],
            )
            .await;
        info!("rows: {:?}", rows);

        if let Ok(msg) = rows {
            info!("msg: {:?}", msg);
            let rows = client
                .execute(
                    "INSERT INTO publishable_models 
                        VALUES ('DEU-Germany', 'LVA-Latvia', 'study', true),
                                ('DEU-Germany', 'GRC-Greece', 'study', true),
                                ('DEU-Germany', 'LVA-Latvia', 'work', true),
                                ('DEU-Germany', 'GRC-Greece', 'work', true),
                                ('LVA-Latvia', 'DEU-Germany', 'study', true),
                                ('LVA-Latvia', 'GRC-Greece', 'study', true),
                                ('LVA-Latvia', 'DEU-Germany', 'work', true),
                                ('LVA-Latvia', 'GRC-Greece', 'work', true),
                                ('GRC-Greece', 'DEU-Germany', 'study', true),
                                ('GRC-Greece', 'LVA-Latvia', 'study', true),
                                ('GRC-Greece', 'DEU-Germany', 'work', true),
                                ('GRC-Greece', 'LVA-Latvia', 'work', true);",
                    &[],
                )
                .await;
            info!("rows: {:?}", rows);

            if let Ok(msg) = rows {
                info!("msg: {:?}", msg);
                info!("Log Result Ok");
                return Ok(warp::reply::with_status(ok_message, StatusCode::OK));
            } else {
                info!("Log Result Error");
                return Ok(warp::reply::with_status(
                    internal_server_error_message,
                    StatusCode::INTERNAL_SERVER_ERROR,
                ));
            }
        } else {
            info!("Log Result Error");
            return Ok(warp::reply::with_status(
                internal_server_error_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}

pub async fn get_all_user_journeys_handler(_tail: String) -> Result<impl warp::Reply, Infallible> {
    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());

    // connect to DB
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(_e) = connection.await {
                info!("Log Connection Error");
            }
        });

        // read all user journeys from db
        let rows = client
            .query("SELECT id, title, drawio_diagram, uj_type,
                    user_journey_models.workflow_type AS workflow_type,
                    user_journey_models.origin_country AS origin_country,
                    user_journey_models.destination_country AS destination_country, 
                    status, trust_indicator, engine_id, json_description, timestamp,
                    publishable
                FROM user_journey_models 
                LEFT JOIN publishable_models 
                    ON user_journey_models.origin_country = publishable_models.origin_country
                    AND user_journey_models.destination_country = publishable_models.destination_country
                    AND user_journey_models.workflow_type = publishable_models.workflow_type
                    AND uj_type = 'concrete'
                ORDER BY user_journey_models.id;", &[])
            .await;
        info!("rows: {:?}", rows);

        if let Ok(msg) = rows {
            info!("msg: {:?}", msg);
            let entries: Vec<messages::StorageEntry> = msg
                .into_iter()
                .map(|row| {
                    return messages::StorageEntry::new(
                        row.get("id"),
                        row.get("title"),
                        row.get("drawio_diagram"),
                        row.get("uj_type"),
                        row.get("workflow_type"),
                        row.get("origin_country"),
                        row.get("destination_country"),
                        row.get("status"),
                        row.get("trust_indicator"),
                        row.get("engine_id"),
                        row.get("json_description"),
                        row.get("timestamp"),
                        row.get("publishable"),
                    );
                })
                .collect();
            info!("entries: {:?}", entries);

            // create and send response
            let response = messages::AllUserJourneysResponse::new(
                StatusCode::OK.as_u16(),
                "OK".into(),
                entries,
            );

            info!("response: {:?}", response);
            return Ok(warp::reply::with_status(
                warp::reply::json(&response),
                StatusCode::OK,
            ));
        } else {
            info!("Log Result Error");
            return Ok(warp::reply::with_status(
                internal_server_error_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}

pub async fn get_user_journey_handler(
    user_journey_id: String,
    param: std::collections::HashMap<String, String>,
    _tail: String,
) -> Result<impl warp::Reply, Infallible> {
    info!("user_journey_id: {:?}", user_journey_id);

    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());
    let bad_request_message = warp::reply::json(&messages::StatusMessage::bad_request());
    let not_found_message = warp::reply::json(&messages::StatusMessage::not_found());

    // parse id
    let id: Result<Uuid, uuid::Error> = Uuid::parse_str(user_journey_id.as_str());
    //let id = user_journey_id.parse::<i64>();
    if id.is_err() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let id = id.unwrap();
    info!("id: {:?}", id);

    let timestamp = param.get("timestamp");
    if timestamp.is_none() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let timestamp = timestamp.unwrap();
    info!("timestamp: {:?}", timestamp);

    // connect db
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(_e) = connection.await {
                info!("Log Connection Error");
            }
        });

        // read user journey from db
        let rows = client
            .query(
                "SELECT title, drawio_diagram
                    FROM user_journey_models
                    WHERE timestamp=$1 AND id=$2;",
                &[timestamp, &id],
            )
            .await;
        info!("rows: {:?}", rows);

        if let Ok(msg) = rows {
            info!("msg: {:?}", msg);
            if msg.len() > 1 {
                // unexpected result
                info!("Log Result Error (msg.len() > 1)");
                return Ok(warp::reply::with_status(
                    internal_server_error_message,
                    StatusCode::INTERNAL_SERVER_ERROR,
                ));
            } else {
                if let Some(row) = msg.get(0) {
                    info!("row: {:?}", row);
                    let title: String = row.get("title");
                    info!("title: {}", title);
                    let drawio_diagram: String = row.get("drawio_diagram");
                    info!("drawio_diagram: {}", drawio_diagram);

                    // create and send response
                    let response: messages::UserJourneyResponse = messages::UserJourneyResponse::new(
                        StatusCode::OK.as_u16(),
                        "OK".into(),
                        title,
                        drawio_diagram,
                    );
                    info!("Log Result Ok");
                    return Ok(warp::reply::with_status(
                        warp::reply::json(&response),
                        StatusCode::OK,
                    ));
                } else {
                    // not found
                    info!("Log Result Not Found");
                    return Ok(warp::reply::with_status(
                        not_found_message,
                        StatusCode::NOT_FOUND,
                    ));
                }
            }
        } else {
            info!("Log Result Error");
            return Ok(warp::reply::with_status(
                internal_server_error_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}

pub async fn store_user_journey_handler(
    _tail: String,
    body: warp::hyper::body::Bytes,
) -> Result<impl warp::Reply, Infallible> {
    let bad_request_message = warp::reply::json(&messages::StatusMessage::bad_request());
    let created_message = warp::reply::json(&messages::StatusMessage::created());
    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());

    let body_string = String::from_utf8(body.to_vec());
    if body_string.is_err() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_string = body_string.unwrap_or_default();
    info!("Log Request with"; "body" => body_string.clone());

    let body_json_res = serde_json::from_str(&body_string);
    if body_json_res.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    // create storage entry
    let body_json: messages::StoreUJMessage = body_json_res.unwrap();
    info!("body_json: {:?}", body_json);

    let drawio_diagram = body_json.get_drawio_diagram();

    // connect to DB
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(_e) = connection.await {
                info!("Log Connection Error");
            }
        });

        // generate user journey from xml
        let user_journey = usecases::generate_user_journey_from_drawio(drawio_diagram.clone());
        if user_journey.is_ok() {
            let user_journey = user_journey.unwrap();
            let workflow_type = user_journey.get_workflow_type(); // can be none
            let origin_country = user_journey.get_origin_country(); // can be none
            let destination_country = user_journey.get_destination_country(); // can be none

            // get user journey data from json body
            let title = body_json.get_title();
            let timestamp = body_json.get_timestamp();

            // set default data
            let status = "unpublished".to_string();
            let trust_indicator: Option<bool> = None;
            let engine_id: Option<String> = None;

            // init mutable data for abstract case
            let mut uj_type = "abstract".to_string();
            let mut json_description: Option<String> = None;

            if user_journey.is_concrete() {
                uj_type = "concrete".to_string();

                let user_journey_string =
                    usecases::generate_user_journey_string_from_drawio(drawio_diagram.clone());
                if user_journey_string.is_ok() {
                    json_description = Some(user_journey_string.unwrap());
                } else {
                    // should be created if uj is concrete
                    info!("Log BAD REQUEST uj string cannot be created");
                    return Ok(warp::reply::with_status(
                        bad_request_message,
                        StatusCode::BAD_REQUEST,
                    ));
                }
            }

            // store user journey data
            let rows = client
                .execute(
                    "INSERT INTO user_journey_models (title, drawio_diagram,
                            uj_type, workflow_type, origin_country, destination_country,
                            status, trust_indicator, engine_id, json_description, timestamp)
                        VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);",
                    &[
                        &title,
                        &drawio_diagram,
                        &uj_type,
                        &workflow_type,
                        &origin_country,
                        &destination_country,
                        &status,
                        &trust_indicator,
                        &engine_id,
                        &json_description,
                        &timestamp,
                    ],
                )
                .await;
            info!("rows: {:?}", rows);

            if let Ok(msg) = rows {
                info!("Log Result Created");
                info!("msg id: {:?}", msg);

                return Ok(warp::reply::with_status(
                    created_message,
                    StatusCode::CREATED,
                ));
            } else {
                info!("Log Result Error");
                return Ok(warp::reply::with_status(
                    internal_server_error_message,
                    StatusCode::INTERNAL_SERVER_ERROR,
                ));
            }
        } else {
            info!("Log BAD REQUEST uj cannot be created from xml");
            return Ok(warp::reply::with_status(
                bad_request_message,
                StatusCode::BAD_REQUEST,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}

pub async fn update_user_journey_handler(
    user_journey_id: String,
    _tail: String,
    body: warp::hyper::body::Bytes,
) -> Result<impl warp::Reply, Infallible> {
    let bad_request_message = warp::reply::json(&messages::StatusMessage::bad_request());
    let not_found_message = warp::reply::json(&messages::StatusMessage::not_found());
    let ok_message = warp::reply::json(&messages::StatusMessage::ok());
    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());

    // read json body as string from bytes
    let body_string = String::from_utf8(body.to_vec());
    if body_string.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_string = body_string.unwrap_or_default();
    info!("Log Request with"; "body" => body_string.clone());
    let body_json_res = serde_json::from_str(&body_string);
    let body_json: serde_json::Value = body_json_res.unwrap();
    info!("body_json: {:?}", body_json);

    // get approvement value from json
    let approved = &body_json["trust_indicator"];
    info!("approved: {:?}", approved);

    // get timestamp value from json
    let timestamp = &body_json["timestamp"];
    info!("timestamp: {:?}", timestamp);

    // get engine_id value from json
    let engine_id = &body_json["engine_id"];
    info!("engine_id: {:?}", engine_id);

    // parse id
    let id: Result<Uuid, uuid::Error> = Uuid::parse_str(user_journey_id.as_str());
    //let id = user_journey_id.parse::<i64>();
    if id.is_err() || engine_id.is_null() || timestamp.is_null() || approved.is_null() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }

    let id = id.unwrap();
    info!("id: {:?}", id);

    // connect to DB
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(_e) = connection.await {
                info!("Log Connection Error");
            }
        });

        // update user journey approvement in db
        let rows = client
            .execute(
                "UPDATE user_journey_models
                    SET trust_indicator=$1
                    WHERE timestamp=$2 AND status='published' AND id=$3;",
                &[&approved.as_bool(), &timestamp.as_str(), &id],
            )
            .await;
        info!("rows: {:?}", rows);

        // send to engine

        if let Ok(msg) = rows {
            if msg >= 1 {
                info!("msg: {:?}", msg);

                // create engine url
                let url =
                    reqwest::Url::parse_with_params(
                        &format!("{}/{}", config::ENGINE_URL, engine_id.as_str().unwrap()),
                        &[("trust", approved.to_string())]
                    );
                info!("url: {:?}", url);

                if url.is_err() {
                    info!("Log BAD REQUEST");
                    return Ok(warp::reply::with_status(
                        bad_request_message,
                        StatusCode::BAD_REQUEST,
                    ));
                }

                let engine_client = reqwest::Client::new();
                let res = engine_client.put(url.clone().unwrap()).send().await;        

                if res.is_err() {
                    info!("Log Result Error");
                    let err_status = res.unwrap_err().status().unwrap_or_default();
                    let status_message =
                        warp::reply::json(&messages::StatusMessage::new(err_status));
                    return Ok(warp::reply::with_status(status_message, err_status));
                } else {
                    info!("Log Result Ok");
                    let mut info_msg = messages::StatusMessage::ok();
                    let url = url.unwrap();
                    let url_str = url.as_str();
                    info_msg.message = format!("OK url: {:?}; engine_id: {}, as string: {}", url, engine_id.as_str().unwrap(), url_str);
                    let info_message = warp::reply::json(&info_msg);
                    return Ok(warp::reply::with_status(info_message, StatusCode::OK));
                }
            } else {
                info!("Log Not Found");
                info!("msg: {:?}", msg);
                return Ok(warp::reply::with_status(
                    not_found_message,
                    StatusCode::NOT_FOUND,
                ));
            }
        } else {
            info!("Log Result Error");
            return Ok(warp::reply::with_status(
                internal_server_error_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}

pub async fn delete_user_journey_handler(
    user_journey_id: String,
    _tail: String,
    body: warp::hyper::body::Bytes,
) -> Result<impl warp::Reply, Infallible> {
    let bad_request_message = warp::reply::json(&messages::StatusMessage::bad_request());
    let ok_message = warp::reply::json(&messages::StatusMessage::ok());
    let internal_server_error_message =
        warp::reply::json(&messages::StatusMessage::internal_server_error());
    let not_found_message = warp::reply::json(&messages::StatusMessage::not_found());

    // parse id
    let id: Result<Uuid, uuid::Error> = Uuid::parse_str(user_journey_id.as_str());
    //let id = user_journey_id.parse::<i64>();
    if id.is_err() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let id = id.unwrap();
    info!("delete id: {:?}", id);

    // read json body as string from bytes
    let body_string = String::from_utf8(body.to_vec());
    if body_string.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_string = body_string.unwrap_or_default();
    info!("Log Request with"; "body" => body_string.clone());
    let body_json_res = serde_json::from_str(&body_string);
    if body_json_res.is_err() {
        info!("Log Request (BAD REQUEST)");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let body_json: serde_json::Value = body_json_res.unwrap();
    info!("body_json: {:?}", body_json);

    let timestamp: Option<&str> = body_json["timestamp"].as_str();
    if timestamp.is_none() {
        info!("Log BAD REQUEST");
        return Ok(warp::reply::with_status(
            bad_request_message,
            StatusCode::BAD_REQUEST,
        ));
    }
    let timestamp: String = timestamp.unwrap().to_string();
    info!("timestamp: {:?}", timestamp);

    // connect to DB
    if let Ok((client, connection)) = tokio_postgres::connect(
        &config::DATABASE_URL
            .to_owned(),
        NoTls,
    )
    .await
    {
        tokio::spawn(async move {
            if let Err(_e) = connection.await {
                info!("Log Connection Error");
            }
        });

        // delete user journey if conditions are met
        let rows = client
            .execute(
                "DELETE FROM user_journey_models
                    WHERE timestamp=$1 AND (status='deactivated' OR status='unpublished') AND id=$2;",
                &[&timestamp, &id],
            )
            .await;

        if let Ok(msg) = rows {
            if msg >= 1 {
                info!("Log Result Ok");
                info!("msg: {:?}", msg);
                return Ok(warp::reply::with_status(ok_message, StatusCode::OK));
            } else {
                info!("Log Not Found");
                info!("msg: {:?}", msg);
                return Ok(warp::reply::with_status(
                    not_found_message,
                    StatusCode::NOT_FOUND,
                ));
            }
        } else {
            info!("Log Connection Error");
            return Ok(warp::reply::with_status(
                internal_server_error_message,
                StatusCode::INTERNAL_SERVER_ERROR,
            ));
        }
    } else {
        info!("Log Connection Error");
        return Ok(warp::reply::with_status(
            internal_server_error_message,
            StatusCode::INTERNAL_SERVER_ERROR,
        ));
    }
}
