use super::UseCaseExecError;
use crate::entities::{drawio_xml, workflow::{Workflow, Action}};
use std::collections::HashMap;

pub fn from_drawio_xml(mx_graph_model: &drawio_xml::MxGraphModel) -> Result<Workflow, UseCaseExecError> {

    let mut workflow = Workflow::new();
    let actions_by_parent = mx_graph_model.get_actions_by_parent_map();
    let edges_by_source = mx_graph_model.get_edge_mx_cells_ref_by_source_map();
    let phase_by_id = mx_graph_model.get_phase_by_id_map();
    let touchpoints_by_parent = mx_graph_model.get_touchpoints_by_parent_map();
    let obejcts_by_parent = mx_graph_model.get_obejcts_by_parent_map();

    // get start phase
    let start_phases = mx_graph_model.get_start_phases();
    let start_phase = start_phases.get(0).unwrap(); // start_phases must have one phase if graph is valid.
    let start_phase_id = start_phase.get_id();

    set_actions(start_phase_id, &mut workflow, &actions_by_parent, &edges_by_source,
            &phase_by_id, &touchpoints_by_parent, &obejcts_by_parent);

    // try to read properties from user objects
    if let Some(user_objects) = mx_graph_model.get_user_objects() {
        for user_object in user_objects {
            if user_object.has_workflow_properties()
            {
                workflow.origin_country = user_object.get_origin_country().unwrap_or_default();
                workflow.destination_country = user_object.get_destination_country().unwrap_or_default();
                workflow.workflow_type = user_object.get_workflow_type().unwrap_or_default();
                break;
            }
        }
    }
    return Ok(workflow);
}

fn set_actions(phase_id: String, workflow: &mut Workflow, 
    actions_by_parent: &HashMap<String, Vec<&drawio_xml::datatypes::Action>>,
    edges_by_source: &HashMap<String, &drawio_xml::datatypes::MxCell>,
    phase_by_id: &HashMap<String, &drawio_xml::datatypes::Phase>,
    touchpoints_by_parent: &HashMap<String, &drawio_xml::datatypes::Touchpoints>,
    obejcts_by_parent: &HashMap<String, Vec<&drawio_xml::datatypes::Obejct>>) {

    // get actions
    for phase_action in actions_by_parent.get(&phase_id).unwrap() {

        let mut action = Action::new_initial(phase_action.get_id(), phase_action.get_obj_title());

        // get touchpoints
        let action_touchpoint = touchpoints_by_parent.get(&phase_action.get_id()).unwrap();

        // get obejcts/services
        if let Some(obejcts) = obejcts_by_parent.get(&action_touchpoint.get_id()) {
            for action_obejct in obejcts {
                action.add_service_id(action_obejct.get_service_id().unwrap());
            }
        }
        workflow.add_action(action);
    }

    // get next phases
    if let Some(phase_edge) = edges_by_source.get(&phase_id) {
        if let Some(phase_edge_target) = phase_edge.get_target() {
            let next_phase = phase_by_id.get(&phase_edge_target).unwrap();
            set_actions(next_phase.get_id(), workflow, actions_by_parent, edges_by_source,
                phase_by_id, touchpoints_by_parent, obejcts_by_parent)
        }
    }
}
