use crate::entities::bpmn::YaSerialize;

#[derive(Clone, Debug, YaSerialize, PartialEq)]
#[yaserde(rename = "bpmndi:BPMNDiagram")]
pub struct DiBPMNDiagram {
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "bpmndi:BPMNPlane")]
    pub bpmndi_bpmn_plane: DiBPMNPlane
}

impl DiBPMNDiagram {

    pub fn init(process_id: String) -> DiBPMNDiagram {
        DiBPMNDiagram {
            id: format!{"BPMNDiagram_{}", process_id.clone()},
            bpmndi_bpmn_plane: DiBPMNPlane::init(process_id.clone())
        }
    }

    pub fn get_mut_bpmndi_bpmn_plane(&mut self) -> &mut DiBPMNPlane {
        &mut self.bpmndi_bpmn_plane
    }

    pub fn push_to_bpmndi_bpmn_shapes(&mut self, shape: DiBPMNShape) {
        self.bpmndi_bpmn_plane.bpmndi_bpmn_shapes.push(shape)
    }

    pub fn push_to_bpmndi_bpmn_edges(&mut self, edge: DiBPMNEdge) {
        self.bpmndi_bpmn_plane.bpmndi_bpmn_edges.push(edge)
    }

}

#[derive(Clone, Debug, YaSerialize, PartialEq)]
pub struct DiBPMNPlane {
    #[yaserde(attribute, rename = "bpmnElement")]
    pub bpmn_element: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "bpmndi:BPMNShape")]
    pub bpmndi_bpmn_shapes: Vec<DiBPMNShape>,
    #[yaserde(child, rename = "bpmndi:BPMNEdge")]
    pub bpmndi_bpmn_edges: Vec<DiBPMNEdge>
}

impl DiBPMNPlane {

    fn init(process_id: String) -> DiBPMNPlane {
        DiBPMNPlane {
            bpmn_element: process_id.clone(),
            id: format!{"BPMNPlane_{}", process_id.clone()},
            bpmndi_bpmn_shapes: Vec::new(),
            bpmndi_bpmn_edges: Vec::new()
        }
    }
}

#[derive(Clone, Debug, YaSerialize, PartialEq)]
#[yaserde(rename = "bpmndi:BPMNShape")]
pub struct DiBPMNShape {
    #[yaserde(attribute, rename = "bpmnElement")]
    pub bpmn_element: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(attribute, rename = "isExpanded")]
    pub is_expanded: Option<bool>,
    #[yaserde(child, rename = "omgdc:Bounds")]
    pub bounds: DcBounds
}

#[derive(Clone, Debug, YaSerialize, PartialEq)]
pub struct DcBounds {
    #[yaserde(attribute)]
    pub height: String,
    #[yaserde(attribute)]
    pub width: String,
    #[yaserde(attribute)]
    pub x: String,
    #[yaserde(attribute)]
    pub y: String
}

impl DiBPMNShape {

    /** creates initial shape at (x,0) */
    pub fn new_init_event(ref_id: String, x: String) -> DiBPMNShape {
        DiBPMNShape {
            bpmn_element: ref_id.clone(),
            id: format!{"BPMNShape_{}", ref_id.clone()},
            is_expanded: None,
            bounds: DcBounds {
                height: "36.0".to_string(),
                width:  "36.0".to_string(),
                x:      x,
                y:      "0".to_string()
            }
        }
    }

    /** creates initial shape at (x,-14.0) */
    pub fn new_init_task(ref_id: String, x: String) -> DiBPMNShape {
        DiBPMNShape {
            bpmn_element: ref_id.clone(),
            id: format!{"BPMNShape_{}", ref_id.clone()},
            is_expanded: None,
            bounds: DcBounds {
                height: "65.0".to_string(),
                width:  "127.0".to_string(),
                x:      x,
                y:      "-14.0".to_string()
            }
        }
    }

    /** creates initial shape at (x,0) */
    pub fn new_init_sub_process(ref_id: String, x: String, width: String) -> DiBPMNShape {
        DiBPMNShape {
            bpmn_element: ref_id.clone(),
            id: format!{"BPMNShape_{}", ref_id.clone()},
            is_expanded: Some(true),
            bounds: DcBounds {
                height: "210.0".to_string(),
                width:  width,
                x:      x,
                y:      "-85.0".to_string()
            }
        }
    }

    pub fn get_mut_bounds(&mut self) -> &mut DcBounds {
        &mut self.bounds
    }
}


#[derive(Clone, Debug, YaSerialize, PartialEq)]
#[yaserde(rename = "bpmndi:BPMNEdge")]
pub struct DiBPMNEdge {
    #[yaserde(attribute, rename = "bpmnElement")]
    pub bpmn_element: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "omgdi:waypoint")]
    pub waypoints: Vec<DiWaypoint>
}

#[derive(Clone, Debug, YaSerialize, PartialEq)]
pub struct DiWaypoint {
    #[yaserde(attribute)]
    pub x: String,
    #[yaserde(attribute)]
    pub y: String
}

impl DiBPMNEdge {
    /** creates initial Edge from (xs,18) to (xt,18) */
    pub fn init(ref_id: String, xs: String, xt: String) -> DiBPMNEdge {
        DiBPMNEdge {
            bpmn_element:   ref_id.clone(),
            id:             format!{"BPMNEdge_{}", ref_id.clone()},
            waypoints:      vec![
                DiWaypoint {x: xs, y: "18".to_string()},
                DiWaypoint {x: xt, y: "18".to_string()}
            ]
        }
    }
}