use yaserde_derive::YaDeserialize;

#[derive(Debug, Default, YaDeserialize)]
pub struct Object {
    #[yaserde(attribute)]
    pub service_id: Option<String>,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: MxCell
}

impl Object {
    pub fn new() -> Object {
        Object {
            service_id: None,
            mx_cell: MxCell::new(),
        }
    }
}

/**
 *  Represents the cell that stores the UJ Properties 
 *  and the UJ Descriptions.
 */
#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "ujproperties")]
pub struct UJProperties {
    #[yaserde(attribute)]
    pub label: Option<String>,
    #[yaserde(attribute)]
    pub origin_country: Option<String>,
    #[yaserde(attribute)]
    pub destination_country: Option<String>,
    #[yaserde(attribute)]
    pub workflow_type: Option<String>,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: Option<MxCell>,
    #[yaserde(child)]
    pub ujlang: Vec<UjLang>
}

impl UJProperties {

    pub fn has_workflow_properties(&self) -> bool {
        self.origin_country.is_some() 
            && self.destination_country.is_some() 
            && self.workflow_type.is_some()
            && !self.get_origin_country().unwrap_or_default().trim().eq("")
            && !self.get_destination_country().unwrap_or_default().trim().eq("")
            && !self.get_workflow_type().unwrap_or_default().trim().eq("")
    }

    pub fn has_engl_uj_desc(&self) -> bool {
        // TODO
        return false;
    }

    pub fn get_origin_country(&self) -> Option<String> {
        if let Some(origin_country) = &self.origin_country {
            Some(origin_country.to_string())
        } else {
            None
        }
    }

    pub fn get_destination_country(&self) -> Option<String> {
        if let Some(destination_country) = &self.destination_country {
            Some(destination_country.to_string())
        } else {
            None
        }
    }

    pub fn get_workflow_type(&self) -> Option<String> {
        if let Some(workflow_type) = &self.workflow_type {
            Some(workflow_type.to_string())
        } else {
            None
        }
    }

    pub fn get_ujlang_list(&self) -> &Vec<UjLang> {
        &self.ujlang
    }

}

#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "phase")]
pub struct Phase {
    #[yaserde(attribute)]
    pub xmlns: Option<String>,
    #[yaserde(attribute)]
    pub placeholders: String,
    #[yaserde(attribute)]
    pub placeholder: String,
    #[yaserde(attribute)]
    pub obj_title: String, // displayed title
    #[yaserde(attribute)]
    pub label: Option<String>,
    #[yaserde(attribute)]
    pub tooltip: String,
    #[yaserde(attribute)]
    pub ujmt_obj_type: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: MxCell,
    #[yaserde(child)]
    pub ujlang: Vec<UjLang>
}

impl Phase {

    pub fn get_mx_cell_ref(&self) -> &MxCell {
        &self.mx_cell
    }

    pub fn get_id(&self) -> String {
        self.id.to_string()
    }

    pub fn get_obj_title(&self) -> String {
        self.obj_title.to_string()
    }

    pub fn get_ujlang_list(&self) -> &Vec<UjLang> {
        &self.ujlang
    }

}

#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "action")]
pub struct Action {
    #[yaserde(attribute)]
    pub xmlns: Option<String>,
    #[yaserde(attribute)]
    pub placeholders: String,
    #[yaserde(attribute)]
    pub placeholder: String,
    #[yaserde(attribute)]
    pub obj_title: String, // displayed title
    #[yaserde(attribute)]
    pub obj_description: Option<String>, // displayed title
    #[yaserde(attribute)]
    pub tooltip: String,
    #[yaserde(attribute)]
    pub ujmt_obj_type: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: MxCell,
    #[yaserde(child)]
    pub ujlang: Vec<UjLang>
}

impl Action {

    pub fn get_mx_cell_ref(&self) -> &MxCell {
        &self.mx_cell
    }

    pub fn get_id(&self) -> String {
        self.id.to_string()
    }

    pub fn is_mandatory(&self) -> bool {
        self.obj_title.trim().starts_with("★")
    }

    pub fn get_obj_title(&self) -> String {
        self.obj_title.to_string()
    }

    pub fn get_parent(&self) -> Option<String> {
        if let Some(parent) = &self.get_mx_cell_ref().parent {
            Some(parent.to_string())
        } else {
            None
        }
    }

    pub fn get_obj_description(&self) -> Option<String> {
        if let Some(obj_description) = &self.obj_description {
            Some(obj_description.to_string())
        } else {
            None
        }
    }

    pub fn get_ujlang_list(&self) -> &Vec<UjLang> {
        &self.ujlang
    }

}

#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "description")]
pub struct Description {
    #[yaserde(attribute)]
    pub xmlns: Option<String>,
    #[yaserde(attribute)]
    pub placeholders: String,
    #[yaserde(attribute)]
    pub placeholder: String,
    #[yaserde(attribute)]
    pub tooltip: String,
    #[yaserde(attribute)]
    pub ujmt_obj_type: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: MxCell,
    #[yaserde(child)]
    pub ujlang: Vec<UjLang>
}

impl Description {

    pub fn get_ujlang_list(&self) -> &Vec<UjLang> {
        &self.ujlang
    }

}

#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "userinfo")]
pub struct UserInfo {
    #[yaserde(attribute)]
    pub xmlns: Option<String>,
    #[yaserde(attribute)]
    pub placeholders: String,
    #[yaserde(attribute)]
    pub label: String,
    #[yaserde(attribute)]
    pub tooltip: String,
    #[yaserde(attribute)]
    pub ujmt_obj_type: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: MxCell
}

#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "touchpoints")]
pub struct TouchpointList {
    #[yaserde(attribute)]
    pub xmlns: Option<String>,
    #[yaserde(attribute)]
    pub placeholders: String,
    #[yaserde(attribute)]
    pub label: String,
    #[yaserde(attribute)]
    pub tooltip: String,
    #[yaserde(attribute)]
    pub ujmt_obj_type: String,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: MxCell
}

impl TouchpointList {

    pub fn get_mx_cell_ref(&self) -> &MxCell {
        &self.mx_cell
    }

    pub fn get_id(&self) -> String {
        self.id.to_string()
    }

    pub fn get_parent(&self) -> Option<String> {
        if let Some(parent) = &self.get_mx_cell_ref().parent {
            Some(parent.to_string())
        } else {
            None
        }
    }
}

// rename to Service
#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "ServiceObject")]
pub struct Touchpoint {
    #[yaserde(attribute, rename = "serviceID")]
    pub service_id: String,
    #[yaserde(attribute)]
    pub link: String,
    #[yaserde(attribute)]
    pub service_description: Option<String>,
    #[yaserde(attribute)]
    pub location: String,
    #[yaserde(attribute)]
    pub service_title: Option<String>,
    #[yaserde(attribute)]
    pub label: Option<String>,
    #[yaserde(attribute)]
    pub placeholders: Option<String>,
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(child, rename = "mxCell")]
    pub mx_cell: MxCell
}

impl Touchpoint
{    
    pub fn get_service_id(&self) -> String {
        self.service_id.to_string()
    }

    pub fn get_id(&self) -> String {
        self.id.to_string()
    }

    pub fn get_location(&self) -> String {
        self.location.to_string()
    }

    pub fn get_service_description(&self) -> Option<String> {
        if let Some(service_description) = &self.service_description {
            Some(service_description.to_string())
        } else {
            None
        }
    }

    pub fn get_service_title(&self) -> Option<String> {
        if let Some(service_title) = &self.service_title {
            Some(service_title.to_string())
        } else {
            None
        }
    }

    pub fn get_link(&self) -> String {
        self.link.to_string()
    }

    pub fn get_mx_cell_ref(&self) -> &MxCell {
        &self.mx_cell
    }

    pub fn get_parent(&self) -> Option<String> {
        if let Some(parent) = &self.get_mx_cell_ref().parent {
            Some(parent.to_string())
        } else {
            None
        }
    }
}

#[derive(Debug, Default, YaDeserialize)]
//#[yaserde(rename = "mxCell")]
pub struct MxCell {
    #[yaserde(attribute)]
    pub id: String,
    #[yaserde(attribute)]
    pub parent: Option<String>,
    #[yaserde(attribute)]
    pub value: Option<String>,
    #[yaserde(attribute)]
    pub style: Option<String>,
    #[yaserde(attribute)]
    pub vertex: Option<String>,
    #[yaserde(attribute)]
    pub edge: Option<String>,
    #[yaserde(attribute)]
    pub source: Option<String>,
    #[yaserde(attribute)]
    pub target: Option<String>,
    #[yaserde(child, rename = "mxGeometry")]
    pub mx_geometry: Option<MxGeometry>
}

impl MxCell {
    pub fn new() -> MxCell {
        MxCell {
            id: "-1".to_string(),
            parent: None,
            value: None,
            style: None,
            vertex: None,
            edge: None,
            source: None,
            target: None,
            mx_geometry: None,
        }
    }

    pub fn get_id(&self) -> String {
        self.id.to_string()
    }

    pub fn get_parent(&self) -> Option<String> {
        if let Some(parent) = &self.parent {
            Some(parent.to_string())
        } else {
            None
        }
    }

    pub fn get_target(&self) -> Option<String> {
        if let Some(target) = &self.target {
            Some(target.to_string())
        } else {
            None
        }
    }

    pub fn get_source(&self) -> Option<String> {
        if let Some(source) = &self.source {
            Some(source.to_string())
        } else {
            None
        }
    }

    pub fn is_edge(&self) -> bool {
        self.edge.is_some() && self.vertex.is_none()
    }

    pub fn is_vertex(&self) -> bool {
        self.edge.is_none() && self.vertex.is_some()
    }
}

#[derive(Debug, YaDeserialize)]
#[yaserde(rename = "mxGeometry")]
pub struct MxGeometry {
    #[yaserde(attribute)]
    pub x: Option<String>,
    #[yaserde(attribute)]
    pub y: Option<String>,
    #[yaserde(attribute)]
    pub width: Option<String>,
    #[yaserde(attribute)]
    pub height: Option<String>,
    #[yaserde(attribute)]
    pub relative: Option<String>,
    #[yaserde(rename = "as")]
    pub attr_as: Option<String>
}

#[derive(Debug, Default, YaDeserialize)]
#[yaserde(rename = "ujlang")]
pub struct UjLang {
    #[yaserde(attribute)]
    pub language: String,
    #[yaserde(attribute)]
    pub value: String,
}

impl UjLang {

    pub fn get_language(&self) -> String {
        self.language.to_string()
    }

    pub fn get_value(&self) -> String {
        self.value.to_string()
    }
}