use failure::Fail;
use serde::{Serialize, Deserialize};
use warp::http::StatusCode;

use uuid::Uuid;

#[derive(Fail, Debug)]
pub enum MessageError {
    #[fail(display = "The JSON Message cannot be deserialized. ({})", _0)]
    DeserializationFailed(serde_json::error::Error),
    #[fail(display = "The JSON Message cannot be serialized. ({})", _0)]
    SerializationFailed(serde_json::error::Error)
}

/* -------------------------
    Status Communication
------------------------- */

#[derive(Serialize)]
pub struct StatusMessage {
    code: u16,
    pub message: String,
}

impl StatusMessage
{
    pub fn not_found() -> StatusMessage {
        StatusMessage {
            code: StatusCode::NOT_FOUND.as_u16(),
            message: "NOT FOUND".into()
        }
    }

    pub fn ok() -> StatusMessage {
        StatusMessage {
            code: StatusCode::OK.as_u16(),
            message: "OK".into()
        }
    }

    pub fn created() -> StatusMessage {
        StatusMessage {
            code: StatusCode::CREATED.as_u16(),
            message: "CREATED".into()
        }
    }

    pub fn forbidden() -> StatusMessage {
        StatusMessage {
            code: StatusCode::FORBIDDEN.as_u16(),
            message: "FORBIDDEN".into()
        }
    }

    pub fn bad_request() -> StatusMessage {
        StatusMessage {
            code: StatusCode::BAD_REQUEST.as_u16(),
            message: "BAD REQUEST".into()
        }
    }

    pub fn internal_server_error() -> StatusMessage {
        StatusMessage {
            code: StatusCode::INTERNAL_SERVER_ERROR.as_u16(),
            message: "INTERNAL SERVER ERROR".into()
        }
    }

    pub fn new(status: http::StatusCode) -> StatusMessage {
        StatusMessage {
            code: status.as_u16(),
            message: status.to_string().into()
        }
    }
}

/* -------------------------
    Mark Response
------------------------- */

#[derive(Debug, Serialize, Deserialize)]
pub struct MarkResponse {
    code: u16,
    message: String,
    concrete: bool,
    touchpoints: Vec<DrawioObject>,
    actions: Vec<DrawioObject>
}

impl MarkResponse {
    pub fn new(status: http::StatusCode, concrete: bool, touchpoints: Vec<DrawioObject>, actions: Vec<DrawioObject>) -> MarkResponse {
        MarkResponse {
            code: status.as_u16(),
            message: status.to_string().into(),
            concrete: concrete,
            touchpoints: touchpoints,
            actions: actions
        }
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub struct DrawioObject {
    drawio_id: String,
    pos: u32,
    // status
}

impl DrawioObject {
    pub fn new(drawio_id: String, pos: u32) -> DrawioObject {
        DrawioObject {
            drawio_id: drawio_id,
            pos: pos
        }
    }
}

/* -------------------------
    Engine Communication
------------------------- */

// Message Format sent to the Engine
/*#[derive(Serialize)]
pub struct UJSERequest {
    #[serde(rename = "destinationCountry")]
    description: String,
    #[serde(rename = "originCountry")]
    origin_country: String,
    #[serde(rename = "destinationCountry")]
    destination_country: String,
    #[serde(rename = "workflowType")]
    workflow_type: String,
    #[serde(rename = "BPMNContent")]
    bpmn_content: String
}

impl UJSERequest {
    pub fn new(origin_country: String, destination_country: String, workflow_type: String, description: String, bpmn_content: String) -> UJSERequest {
        UJSERequest {
            description: description,
            origin_country: origin_country,
            destination_country: destination_country,
            workflow_type: workflow_type,
            bpmn_content: bpmn_content,
        }
    }

    pub fn get_description(&self) -> String {
        self.description.to_string()
    }

    pub fn get_origin_country(&self) -> String {
        self.origin_country.to_string()
    }

    pub fn get_destination_country(&self) -> String {
        self.destination_country.to_string()
    }

    pub fn get_workflow_type(&self) -> String {
        self.workflow_type.to_string()
    }

    pub fn get_bpmn_content(&self) -> String {
        self.bpmn_content.to_string()
    }
}*/

// Workflow Format received from the Engine
#[derive(Deserialize, Serialize)]
pub struct UJSEResponse {
    #[serde(rename = "workflowId")]
    workflow_id: String,
    #[serde(rename = "workflowType")]
    workflow_type: String,
    #[serde(rename = "countryOrigin")]
    country_origin: String,
    #[serde(rename = "countryDestination")]
    country_destination: String,
    jsonfile: String,
    #[serde(rename = "jsonfileInternal")]
    jsonfile_internal: String,
    #[serde(rename = "jsonfileName")]
    jsonfile_name: String
}

// Message Format to return a list of Workflows received from the Engine
#[derive(Serialize)]
pub struct AllWorkflowsUJSEResponse {
    code: u16,
    message: String,
    workflows: Vec<UJSEResponse>
}

impl AllWorkflowsUJSEResponse {
    pub fn new(status: http::StatusCode, workflows: Vec<UJSEResponse>) -> AllWorkflowsUJSEResponse {
        AllWorkflowsUJSEResponse {
            code: status.as_u16(),
            message: status.to_string().into(),
            workflows: workflows
        }
    }
}

/* -------------------------
    Storage Communication
------------------------- */

// Format to store a User Journey in the Storage
#[derive(Debug, Deserialize, Serialize)]
pub struct StoreUJMessage {
    title: String,
    drawio_diagram: String,
    timestamp: String
}

impl StoreUJMessage {
    pub fn new(title: String, drawio_diagram: String, timestamp: String) -> StoreUJMessage
    {
        StoreUJMessage
        {
            title: title,
            drawio_diagram: drawio_diagram,
            timestamp: timestamp
        }
    }

    pub fn get_title(&self) -> String {
        self.title.to_string()
    }

    pub fn get_drawio_diagram(&self) -> String {
        self.drawio_diagram.to_string()
    }

    pub fn get_timestamp(&self) -> String {
        self.timestamp.to_string()
    }

}

// Format to store a User Journey in the Storage
#[derive(Debug, Deserialize, Serialize)]
pub struct StorageEntry {
    id: Uuid,
    title: String,
    drawio_diagram: String,
    uj_type: String,
    workflow_type: Option<String>,
    origin_country: Option<String>,
    destination_country: Option<String>,
    status: String,
    trust_indicator: Option<bool>,
    engine_id: Option<String>,
    json_description: Option<String>,
    timestamp: String,
    publishable: Option<bool>,
}

impl StorageEntry {
    pub fn new(id: Uuid, title: String, drawio_diagram: String, uj_type: String, workflow_type: Option<String>, origin_country: Option<String>,
        destination_country: Option<String>, status: String, trust_indicator: Option<bool>, engine_id: Option<String>, json_description: Option<String>,
        timestamp: String, publishable: Option<bool>) -> StorageEntry
    {
        StorageEntry
        {
            id: id,
            title: title,
            drawio_diagram: drawio_diagram,
            uj_type: uj_type,
            workflow_type: workflow_type,
            origin_country: origin_country,
            destination_country: destination_country,
            status: status,
            trust_indicator: trust_indicator,
            engine_id: engine_id,
            json_description: json_description,
            timestamp: timestamp,
            publishable: publishable
        }
    }

    pub fn get_title(&self) -> String {
        self.title.to_string()
    }

    pub fn get_drawio_diagram(&self) -> String {
        self.drawio_diagram.to_string()
    }

    pub fn get_uj_type(&self) -> String {
        self.uj_type.to_string()
    }

    pub fn get_workflow_type(&self) -> Option<String> {
        if let Some(workflow_type) = &self.workflow_type {
            return Some(workflow_type.to_string());
        } else {
            return None;
        }
    }

    pub fn get_origin_country(&self) -> Option<String> {
        if let Some(origin_country) = &self.origin_country {
            return Some(origin_country.to_string());
        } else {
            return None;
        }
    }

    pub fn get_destination_country(&self) -> Option<String> {
        if let Some(destination_country) = &self.destination_country {
            return Some(destination_country.to_string());
        } else {
            return None;
        }
    }

    pub fn get_status(&self) -> String {
        self.status.to_string()
    }

    pub fn get_trust_indicator(&self) -> Option<bool> {
        return self.trust_indicator;
    }

    pub fn get_engine_id(&self) -> Option<String> {
        if let Some(engine_id) = &self.engine_id {
            return Some(engine_id.to_string());
        } else {
            return None;
        }
    }

    pub fn get_json_description(&self) -> Option<String> {
        if let Some(json_description) = &self.json_description {
            return Some(json_description.to_string());
        } else {
            return None;
        }
    }

    pub fn get_timestamp(&self) -> String {
        self.timestamp.to_string()
    }

}

// Message Format to return a User Journey received from the Storage
#[derive(Serialize)]
pub struct UserJourneyResponse {
    code: u16,
    message: String,
    title: String,
    drawio_diagram: String
}

impl UserJourneyResponse {
    pub fn new(code: u16, message: String, title: String, drawio_diagram: String) -> UserJourneyResponse {
        UserJourneyResponse {
            code: code,
            message: message,
            title: title,
            drawio_diagram: drawio_diagram
        }
    }
}

// Message Format to return all User Journeys received from the Storage
#[derive(Serialize, Debug)]
pub struct AllUserJourneysResponse {
    code: u16,
    message: String,
    entries: Vec<StorageEntry>
}

impl AllUserJourneysResponse {
    pub fn new(code: u16, message: String, entries: Vec<StorageEntry>) -> AllUserJourneysResponse {
        AllUserJourneysResponse {
            code: code,
            message: message,
            entries: entries
        }
    }
}